from typing import Any
import pyqtgraph
import numpy as np
import rasterio.windows
import datetime
import math

from .__prelude__ import (
    logger, Qt, PointData
)
from .FitCurveItem import FitCurveItem

def set_all_pens(item, color, width=0):
    pen = Qt.QPen(color)
    pen.setWidth(width)
    pen.setCosmetic(True)
    item.setPen(pen)
    item.setSymbolPen(pen)
    item.setBrush(color)
    item.setSymbolBrush(color)

class PointDataItem(pyqtgraph.PlotItem):
    def __init__(self, point_data, reference_data):
        super().__init__()
        self.__point_data = point_data
        self.__point_data.fieldChanged.connect(self.__on_point_data_changed)
        self.__point_data.dynamic_attribute("seasonal_period").value_changed.connect(self.__on_period_changed)
        self.__reference_data = reference_data
        self.__reference_data.fieldChanged.connect(self.__on_reference_data_changed)

        self.__selected_band = point_data.selected_band

        self.__fit_item = FitCurveItem(point_data.point_curve.full_curve)
        point_data.point_curve.fieldChanged.connect(self.__on_fit_changed)
        def on_finished(est):
            curve = point_data.point_curve.full_curve
            param_map = {param.name: value for param, value in zip(curve.params, est._current_params)}
            if "step" in param_map:
                step_timestamp = self.fit_item.to_x(param_map["step"])
                step_date = datetime.datetime.fromtimestamp(step_timestamp).strftime('%Y-%m-%d')
                logger.debug("Finished params. Seism at %s", step_date)
        self.fit_item.finished_estimate.connect(on_finished)

        self.__error_bar_low = pyqtgraph.PlotCurveItem(connect='finite', pen = point_data.color)
        self.__error_bar_high = pyqtgraph.PlotCurveItem(connect='finite', pen = point_data.color)
        self.__error_bar_item = pyqtgraph.FillBetweenItem(self.__error_bar_low, self.__error_bar_high, brush = point_data.color)
        self.error_bar_item.setZValue(-1.0)
        self.error_bar_item.setVisible(point_data.show_variance)

        self.__scatter_item = pyqtgraph.PlotDataItem(symbolPen = point_data.color, pen = point_data.color, symbolBrush = point_data.color, brush = point_data.color)
        self.scatter_item.scatter.setData(hoverable = True)
        self.scatter_item.sigPointsClicked.connect(self.__on_points_clicked)

        self.addItem(self.scatter_item)
        self.addItem(self.fit_item)
        self.addItem(self.error_bar_item)

        self.scatter_item.setVisible(point_data.show_in_plot)
        self.fit_item.setVisible(point_data.show_in_plot)
        self.__set_color(point_data.color)
        self.__update_plot()
        self.__on_period_changed(1.0)

    @property
    def sigPointsHovered(self):
        return self.scatter_item.sigPointsHovered

    @property
    def scatter_item(self):
        return self.__scatter_item
    @property
    def fit_item(self):
        return self.__fit_item
    @property
    def error_bar_item(self):
        return self.__error_bar_item
    @property
    def point_data(self):
        return self.__point_data

    def __set_color(self, color):
        set_all_pens(self.scatter_item, color)
        set_all_pens(self.fit_item, color, width=3)
        muted_color = Qt.QColor(color)
        muted_color.setAlpha(96)
        self.error_bar_item.setBrush(muted_color)

    def __update_plot(self):
        xs, ys = self.__point_data.point_values
        _, ys_ref = self.__reference_data.point_values
        ys_adjusted = ys-ys_ref
        if xs.shape == ys_adjusted.shape:
            self.scatter_item.setData(x = xs, y = ys_adjusted, connect='finite')
            height = np.sqrt(self.__point_data.point_variance)
            has_data = ~np.isnan(ys_adjusted)
            self.__error_bar_low.setData(xs[has_data], (ys_adjusted - height)[has_data])
            self.__error_bar_high.setData(xs[has_data], (ys_adjusted + height)[has_data])
            self.fit_item.set_fit_points(xs, ys_adjusted)

            self.__point_data.point_curve.timestamp_to_x = self.fit_item.from_x
            if not self.__point_data.point.point_curve.is_empty:
                self.fit_item.start_fit()

    @Qt.Slot(str, Any)
    def __on_point_data_changed(self, field):
        if field == "show_in_plot":
            self.scatter_item.setVisible(self.__point_data.point.show_in_plot)
            self.fit_item.setVisible(self.__point_data.point.show_in_plot)
            self.error_bar_item.setVisible(self.__point_data.point.show_variance and self.__point_data.point.show_in_plot)
        if field == "show_variance":
            self.error_bar_item.setVisible(self.__point_data.point.show_variance and self.__point_data.point.show_in_plot)
        if field == "color":
            self.__set_color(self.__point_data.point.color)
        if field == "point_values":
            self.__update_plot()

    @Qt.Slot(float)
    def __on_period_changed(self, old):
        p_curve = self.__point_data.point.point_curve
        period = self.__point_data.seasonal_period
        p_curve.seasonal_curve.period = period
        p_curve.semi_seasonal_curve.period = period / 2.0

    @Qt.Slot(str, Any)
    def __on_reference_data_changed(self, field):
        if field == "point_values":
            self.__update_plot()

    @Qt.Slot(str, Any)
    def __on_fit_changed(self, field):
        if field == "full_curve":
            p_curve = self.__point_data.point.point_curve
            self.fit_item.set_curve(p_curve.full_curve)
            if p_curve.is_empty:
                self.fit_item.setData()
            else:
                self.fit_item.start_fit()

    @Qt.Slot(Any, Any)
    def __on_points_clicked(self, _a, points):
        if len(points) > 0:
            self.__selected_band.band_number = int(points[0].index())



import numpy as np

from .__prelude__ import Qt, dynamic, SELF
from .CurveParam import CurveParam
from .Curve import Curve

def sigmoid(x):
    return 1.0 / (1.0 + np.exp(-np.clip(x, a_min = -500.0, a_max = 500.0)))
class SigmoidCurve(Curve):
    step          = dynamic.variable()
    initial_step  = dynamic.external()

    def __init__(self):
        super().__init__()
        self.scale = CurveParam('scale')
        self.scale.min_bound = -1.0
        self.scale.max_bound = 1.0
        self.steepness = CurveParam('steepness')
        self.step = CurveParam('step')

        self._dynamic_initial_step = SELF.step.initial[self]

    __mime_type__ = "application/x-insarviz/SigmoidCurve"

    @property
    def nparams(self):
        return 3
    @property
    def params(self):
        return [self.scale, self.steepness, self.step]

    def func(self, xs, scale, steepness, step):
        return scale * sigmoid(np.exp(steepness) * (xs - step))
    def jacobian(self, xs, scale, steepness, step):
        es = np.exp(steepness)
        sig = sigmoid(es * (xs - step))
        sig_prime = (scale * es) * sig * (1 - sig)

        return np.array([
            sig,
            (xs-step) * sig_prime,
            - sig_prime
        ])

    def hessian_epsilon(self, xs, err, grad, s, t, x0):
        et = np.exp(t)
        x_d = xs-x0
        s0 = sigmoid(et*x_d)
        s1 = s0*(1-s0)
        s2 = s1*(1-2*s0)

        et_x_d = et*x_d

        ds_ds = 0.0
        dt_ds = np.sum(err * (et_x_d*s1))
        dx0_ds = np.sum(err * (-et*s1))
        dt_dt = np.sum(err * ((s*et_x_d)*(s1 + et_x_d*s2)))
        dx0_dt = np.sum(err * (-(s*et)*(s1 + et_x_d*s2)))
        dx0_dx0 = np.sum(err * ((s*et*et)*s2))

        hessian = np.array([
            [ds_ds, dt_ds, dx0_ds],
            [dt_ds, dt_dt, dx0_dt],
            [dx0_ds, dx0_dt, dx0_dx0]
        ])
        return np.dot(hessian @ grad, grad)



from typing import Any
from .__prelude__ import (Qt, color_icon, SELF, EACH)

class ItemChooser(Qt.QComboBox):
    current_item_changed = Qt.Signal(int, Any)

    def __init__(self, items):
        super().__init__()
        self._items = items
        self._item_names = EACH(SELF.name)[items]
        self._item_names.value_changed.connect(self._on_items_changed)
        self._item_colors = EACH(SELF.color)[items]
        self._item_colors.value_changed.connect(self._on_items_changed)

        self.addItem('(none)', None)
        for i, p in enumerate(self._items):
            self.addItem(color_icon(p.color), p.name, i)
        self.currentIndexChanged.connect(self._on_index_changed)
        self.__current_item = None

    @Qt.Slot(int)
    def _on_index_changed(self, index):
        item_data = self.itemData(index)
        if item_data is None or index == -1:
            self.__current_item = None
            self.current_item_changed.emit(-1, None)
        else:
            self.__current_item = self._items[item_data]
            self.current_item_changed.emit(item_data, self._items[item_data])

    @Qt.Slot(int, str)
    def _on_items_changed(self):
        current = self.__current_item
        current_index = None
        self.clear()
        self.addItem('(none)', None)
        for i, item in enumerate(self._items):
            self.addItem(color_icon(item.color), item.name, i)
            if id(item) == id(current):
                current_index = i
        if current_index is not None:
            self.setCurrentIndex(current_index+1)



from .CurveSum         import CurveSum
from .CurveEstimate    import CurveEstimate
from .CurveParam       import CurveParam
from .Curve            import Curve
from .LinearCurve      import LinearCurve, ConstantCurve
from .HeavisideCurve   import HeavisideCurve
from .PeriodicCurve    import PeriodicCurve
from .SigmoidCurve     import SigmoidCurve
from .LogCurve         import LogCurve



class SelectionPoint(ObservableStruct):
    def __init__(self, x, y, r, color):
        self.x = x
        self.y = y
        self.r = r
        self.color = color
        self.show_in_map = True
        self.show_in_plot = True
class SelectionReference(ObservableStruct):
    def __init__(self, x, y, r, color):
        self.x = x
        self.y = y
        self.r = r
        self.color = color
        self.show_in_map = True
class SelectionProfile(ObservableStruct):
    def __init__(self, x0, y0, x1, y1, r, color):
        self.x0 = x0
        self.y0 = y0
        self.x1 = x1
        self.y1 = y1
        self.r = r
        self.color = color
        self.show_in_map = True
        
class Selections(ObservableStruct):
    def __init__(self):
        self.points     = ObservableList()
        self.references = ObservableList()
        self.profiles   = ObservableList()

class Project(ObservableStruct):
    def __init__(self):
        self.layers = ObservableList()
        self.selections = Selections()
        self.reference_band = IndexInto(self.selections.references)




import os
import rasterio
import pathlib
import json
import urllib.request

from .__prelude__ import Qt, WindowState, Dataset, logger, SELF, version, version_tuple, log_messages
from .MainWidget import MainWidget
from .MapManagerWidget import MapManagerWidget, LayerListMenu
from .GLWidget import GLWidget
from .plot import SpatialPlotWidget, TemporalPlotWidget
from .HelpWidget import HelpWidget
from .IconDockWidget import IconDockWidget
from .WidgetTree import Container, Leaf
from .IconLabel import IconLabel
from .ComputedProgressBar import ComputedProgressBar
from .LoggerWidget import LoggerWidget

class ActionOn:
    def __init__(self, name, title, callback, shortcut = None, checkable = False):
        self._name = name
        self._title = title
        self._callback = callback
        self._shortcut = shortcut
        self._checkable = checkable
    def create(self, target):
        ret = Qt.QAction(self._title)
        def f():
            self._callback(target)
        if self._shortcut is not None:
            ret.setShortcut(Qt.QKeySequence(self._shortcut))
        ret.setCheckable(self._checkable)
        ret.triggered.connect(f)
        setattr(target, f"{self._name}_action", ret)
def action(title, **kwargs):
    def wrap(f):
        f.__register_action__ = ActionOn(f.__name__, title, f, **kwargs)
        return f
    return wrap

class SetInterleaveRunner(Qt.QRunnable):
    def __init__(self, src_file, dst_file, on_progress):
        super().__init__()
        self._src_file, self._dst_file = pathlib.Path(src_file), pathlib.Path(dst_file)
        self._on_progress = on_progress

    def _run_convert(self):
        with rasterio.open(self._src_file, 'r') as src:
            prof = src.profile
            prof["interleave"] = "band"
            prof["BIGTIFF"] = "YES"

            with rasterio.open(self._dst_file, 'w', **prof) as dst:
                for band in range(src.count):
                    dst.set_band_description(band+1, src.descriptions[band])

                percent = 0
                windows = [*src.block_windows(1)]

                for i, (ji, window) in enumerate(windows):
                    if i*100 > (percent+1)*len(windows):
                        percent = percent+1
                        yield float(percent)/100.0
                    dst.write(src.read(window = window), window = window)

            yield 1.0

    def run(self):
        for progress in self._run_convert():
            self._on_progress(progress)

def LinkLabel(text, color, href = None):
    ret = Qt.QLabel()
    if href is None:
        ret.setText(text)
    else:
        ret.setText(f"<a href=\"{href}\">{text}</a>")
        ret.setTextInteractionFlags(Qt.Qt.TextBrowserInteraction)
        ret.setOpenExternalLinks(True)
    ret.setStyleSheet(f"QLabel:hover {{ color: {color}; }}")
    return ret

def get_latest_version():
    try:
        versions = urllib.request.urlopen("https://deformvis.gricad-pages.univ-grenoble-alpes.fr/insarviz/latest/versions.list", timeout=1).read()
        versions = versions.decode('utf-8')
        latest = versions.split("\n")[-2]
        my_major, my_minor, *_ = version_tuple
        global_major, global_minor, *_ = latest.split(".")
        if (my_major, my_minor) < (int(global_major), int(global_minor)):
            return latest
        else:
            return None
    except Exception as e:
        return None

class InsarvizWindow(Qt.QMainWindow):
    def __init__(self, win_state: WindowState):
        super().__init__()
        # Register all actions
        for act in self.__class__.__dict__.values():
            if hasattr(act, '__register_action__'):
                act.__register_action__.create(self)

        self.setWindowTitle("InsarViz")
        self.setWindowIcon(Qt.QIcon('insarviz:logo_insarviz.png'))
        self.setWindowState(Qt.Qt.WindowMaximized)

        self._state: WindowState = win_state
        self._state.fieldChanged.connect(self._on_state_change)

        self._map_manager_widget = MapManagerWidget(self._state)
        self._minimap_widget = GLWidget(self._state.minimap_scene)
        def resize_minimap_viewport(__w__, __h__):
            def set_world_to_clip(ctx):
                ctx.map_world_to_clip = self._dataset_widget.world_to_clip
                self._state.minimap_scene.renderChanged.emit()
            self._minimap_widget.modify_initialized_scene(set_world_to_clip)
        self._temporal_plot_widget = TemporalPlotWidget(self._state.current_band(), self._state.points, self._state.profiles)

        self._spatial_plot_widget = SpatialPlotWidget(self._state.current_band(), self._state.points, self._state.profiles)

        self._dataset_widget = MainWidget(win_state)
        self._dataset_widget.resized_viewport.connect(resize_minimap_viewport)

        win_state.dynamic_attribute("hovered_pixel").drive(self._set_temporal_cursor_point)
        self.setCentralWidget(self._dataset_widget)

        self._map_manager_dock = Qt.QDockWidget("Project", self)
        self._map_manager_dock.setWidget(self._map_manager_widget)
        def on_manager_visible(visible):
            self.show_map_manager_action.setChecked(visible)
        self._map_manager_dock.visibilityChanged.connect(on_manager_visible)
        self.addDockWidget(Qt.Qt.DockWidgetArea.LeftDockWidgetArea, self._map_manager_dock)

        self._minimap_dock = Qt.QDockWidget("Minimap", self)
        self._minimap_dock.setWidget(self._minimap_widget)
        def on_minimap_visible(visible):
            self.show_minimap_action.setChecked(visible)
            if visible:
                # Necesary to redraw OpenGL widgets when floating the dock
                self._minimap_widget.update()
        self._minimap_dock.visibilityChanged.connect(on_minimap_visible)
        self.addDockWidget(Qt.Qt.DockWidgetArea.LeftDockWidgetArea, self._minimap_dock)

        self._geo_map_dock = Qt.QDockWidget("Geo Map", self)
        self._geo_map_dock.setVisible(False)
        self._state.dynamic_attribute("geo_scene").drive(self.__set_geo_widget)

        self.setStatusBar(Qt.QStatusBar())
        def fixedHeightWidget(height):
            def ret():
                w = Qt.QWidget()
                w.setFixedHeight(height)
                return w
            return ret

        latest = get_latest_version()
        if latest is not None:
            upgrade_widget = Leaf(LinkLabel, (f"[new version available : {latest}]", "#b6972d", 'https://deformvis.gricad-pages.univ-grenoble-alpes.fr/insarviz/latest/installation.html'))
        else:
            upgrade_widget = Leaf(LinkLabel, ("[up-to-date]", "green", None))

        self.statusBar().addPermanentWidget(Container(Qt.QHBoxLayout, (), [
            Leaf(Qt.QLabel, ("position", ), id="_position_label"),
            (Leaf(Qt.QWidget, ()), {"stretch":1}),
            (Leaf(ComputedProgressBar, (), id = "_compute_band_progress"), {"stretch":1}),
            Leaf(IconLabel, (Qt.QIcon("insarviz:logo_insarviz.png"),)),
            Leaf(Qt.QLabel, (f"InsarViz v{version}", )),
            upgrade_widget
        ], widget_class = fixedHeightWidget(40)).create(self), stretch = 1)
        win_state.dynamic_attribute("position_info").drive(lambda elements: self._position_label.setText(", ".join(elements)))
        self._compute_band_progress.set_computed_value(self._state.current_band().computed_image)

        def on_plots_visible(visible):
            self.show_plot_action.setChecked(self._spatial_plot_dock.isVisible() or self._temporal_plot_dock.isVisible())

        self._temporal_plot_dock = IconDockWidget("Temporal Plots", self, Qt.QIcon("insarviz:temporal.svg"))
        self._temporal_plot_dock.setWidget(self._temporal_plot_widget)
        self._temporal_plot_dock.visibilityChanged.connect(on_plots_visible)

        self._spatial_plot_dock = IconDockWidget("Spatial Plots", self, Qt.QIcon("insarviz:spatial.svg"))
        self._spatial_plot_dock.setWidget(self._spatial_plot_widget)
        self._spatial_plot_dock.visibilityChanged.connect(on_plots_visible)

        self._help_widget = HelpWidget()
        self._help_dialog = Qt.QDialog(self)
        help_layout = Qt.QVBoxLayout(self._help_dialog)
        help_layout.addWidget(self._help_widget)
        self._help_dialog.setLayout(help_layout)
        self._help_dialog.setWindowTitle("InsarViz help")
        self._help_dialog.setModal(True)
        help_shortcut = Qt.QShortcut("F1", self._help_dialog)
        help_shortcut.activated.connect(self._help_dialog.hide)

        self._project_file_path = None

        menu_bar = Qt.QMenuBar()
        file_menu = Qt.QMenu("File")
        file_menu.addAction(self.open_file_action)
        file_menu.addAction(self.save_project_action)
        file_menu.addAction(self.save_project_as_action)
        file_menu.addAction(self.open_geo_lut_action)
        file_menu.addSeparator()
        file_menu.addAction(self.quit_action)
        menu_bar.addMenu(file_menu)

        layer_menu = Qt.QMenu("Layers")
        for act in self._map_manager_widget.layer_actions.actions():
            layer_menu.addAction(act)
        layer_menu.addSeparator()
        self.__add_layer_menu = LayerListMenu(self, self._state.layers)
        for action in self.__add_layer_menu.actions():
            layer_menu.addAction(action)
        menu_bar.addMenu(layer_menu)

        view_menu = Qt.QMenu("View")
        view_menu.addAction(self.show_minimap_action)
        view_menu.addAction(self.show_map_manager_action)
        view_menu.addAction(self.show_plot_action)
        view_menu.addSeparator()
        view_menu.addAction(self.reset_map_camera_action)
        view_menu.addSeparator()
        view_menu.addAction(self.flip_horizontally_action)
        SELF.map_scene.flip_horizontally[self._state].drive(lambda h_flip: self.flip_horizontally_action.setChecked(h_flip))
        view_menu.addAction(self.flip_vertically_action)
        SELF.map_scene.flip_vertically[self._state].drive(lambda v_flip: self.flip_vertically_action.setChecked(v_flip))
        menu_bar.addMenu(view_menu)

        help_menu = Qt.QMenu("Help")
        help_menu.addAction(self.show_help_action)
        help_menu.addAction(self.open_online_doc_action)
        help_menu.addAction(self.show_logs_messages_dialog_action)
        menu_bar.addMenu(help_menu)

        self.setMenuBar(menu_bar)

    @action("Documentation (Web)")
    def open_online_doc(self):
        Qt.QDesktopServices.openUrl('https://deformvis.gricad-pages.univ-grenoble-alpes.fr/insarviz/latest/')

    @action("Reset Map View")
    def reset_map_camera(self):
        self._state.reset_map_camera()

    def _set_temporal_cursor_point(self, pixel_coords):
        w, h = self._state.image_size
        if pixel_coords is None:
            self._temporal_plot_widget.set_cursor_coords(pixel_coords)
            return None
        x, y = pixel_coords
        if x >= 0 and x < w and y >= 0 and y < h:
            self._temporal_plot_widget.set_cursor_coords(pixel_coords)

    def _set_dock_visible(self, dock, visible, area):
        if visible:
            dock.setFloating(False)
            self.addDockWidget(area, dock)
        else:
            dock.setFloating(True)
        dock.setVisible(visible)

    @action("Show Help", shortcut = "F1")
    def show_help(self):
        self._help_dialog.show()

    @action("Show InsarViz logs", shortcut = "F8")
    def show_logs_messages_dialog(self):
        dialog = Qt.QDialog(self)
        dialog.resize(600, 400)
        logger_widget = LoggerWidget(log_messages)
        layout = Qt.QVBoxLayout(dialog)
        layout.addWidget(logger_widget)
        dialog.setModal(True)
        dialog.setWindowTitle("InsarViz logs")
        dialog.setLayout(layout)
        dialog.show()

    @action("Flip Horizontally", checkable=True)
    def flip_horizontally(self):
        checked = self.flip_horizontally_action.isChecked()
        self._state.map_scene.flip_horizontally = checked
    @action("Flip Vertically", checkable=True)
    def flip_vertically(self):
        checked = self.flip_vertically_action.isChecked()
        self._state.map_scene.flip_vertically = checked

    @action("Plot", checkable = True, shortcut = "Ctrl+P")
    def show_plot(self):
        checked = self.show_plot_action.isChecked()
        self._set_dock_visible(self._temporal_plot_dock,
                               checked,
                               Qt.Qt.DockWidgetArea.RightDockWidgetArea)
        self._set_dock_visible(self._spatial_plot_dock,
                               checked,
                               Qt.Qt.DockWidgetArea.RightDockWidgetArea)
        self.tabifyDockWidget(self._spatial_plot_dock, self._temporal_plot_dock)

    @action("Minimap", checkable = True, shortcut = "Ctrl+M")
    def show_minimap(self):
        self._set_dock_visible(self._minimap_dock,
                               self.show_minimap_action.isChecked(),
                               Qt.Qt.DockWidgetArea.LeftDockWidgetArea)
    @action("Project", checkable = True)
    def show_map_manager(self):
        self._set_dock_visible(self._map_manager_dock,
                               self.show_map_manager_action.isChecked(),
                               Qt.Qt.DockWidgetArea.LeftDockWidgetArea)

    def _convert_dataset_interleave(self, path, band_dataset_path):
        progress_dialog = Qt.QDialog()
        progress_dialog.setModal(True)
        progress_dialog.setWindowTitle("Dataset conversion to band interleave")
        layout = Qt.QVBoxLayout()
        label = Qt.QLabel(f"Converting dataset {path}")
        layout.addWidget(label)
        progress_bar = Qt.QProgressBar()
        layout.addWidget(progress_bar)
        progress_dialog.setLayout(layout)
        progress_dialog.setFixedSize(progress_dialog.sizeHint())
        def on_changed():
            if progress_bar.value() > 99:
                progress_dialog.accept()
        progress_bar.valueChanged.connect(on_changed)
        Qt.QThreadPool.globalInstance().start(
            SetInterleaveRunner(path, band_dataset_path,
                                lambda progress: progress_bar.setValue(int(progress*100))))

        ret = progress_dialog.exec()
        return ret == Qt.QDialog.DialogCode.Accepted

    @property
    def __default_open_path(self):
        if self._project_file_path is not None:
            return str(self._project_file_path.parent)
        elif self._state.dataset.file != '<none>':
            return str(self._state.dataset.file.parent)
        else:
            return os.getcwd()

    @action("Load Geo LUT", shortcut = "Ctrl+G")
    def open_geo_lut(self):
        path, filetype = Qt.QFileDialog.getOpenFileName(self, "Open Lookup Table", self.__default_open_path, f"LUT File (*_Lut_*.tiff)")
        if path is None or path == "":
            return
        self._state.geo_lut_dataset = Dataset(pathlib.Path(path))

    def __set_geo_widget(self, geo_scene):
        if geo_scene is None:
            return

        geo_widget = GLWidget(geo_scene)

        def on_position_changed(x_model, y_model):
            x_img, y_img = geo_scene.model_to_image.transform_point((x_model, y_model))
            img = geo_scene.lut_image
            h,w,_ = img.shape
            if x_img >= 0 and x_img <= 1 and y_img >= 0 and y_img <= 1:
                radar_x, radar_y = img[int(y_img*h), int(x_img*w)] * geo_scene.lut_scale
                self._state.map_scene.geo_focus = radar_x, radar_y
            else:
                self._state.map_scene.geo_focus = None
        geo_widget.positionChanged.connect(on_position_changed)

        self._geo_map_dock.setWidget(geo_widget)
        self.addDockWidget(Qt.Qt.DockWidgetArea.RightDockWidgetArea, self._geo_map_dock)
        self._geo_map_dock.setVisible(True)

    @action("Open file", shortcut = "Ctrl+O")
    def open_file(self, path = None):
        DATACUBE_FILES = "Data Cube File (*.tiff *.tif *.h5 *.r4 depl_cumule*)"
        PROJECT_FILES = "InsarViz Project (*.invz)"

        if path is None:
            path, filetype = Qt.QFileDialog.getOpenFileName(self, "Open Data Cube", self.__default_open_path, f"{DATACUBE_FILES};;{PROJECT_FILES};;All files (*)")
            is_dataset = filetype != PROJECT_FILES
        else:
            match path.suffix:
                case ".invz":
                    is_dataset = False
                case _:
                    is_dataset = True

        if path == "":
            return

        if is_dataset:
            path = pathlib.Path(path)
            dataset = Dataset(path)
            if dataset.interleaving != Dataset.Interleaving.band:
                band_dataset_path = path.with_name(path.stem + "_band" + path.suffix)
                if not band_dataset_path.exists():
                    msgBox = Qt.QMessageBox(Qt.QMessageBox.Icon.Warning,
                                            "Suboptimal Interleaving",
                                            f"""
                                            <p>
                                              The dataset at "{path}" has non-band interleaving, which will lead to
                                              (very) slow band loads and computations. It is advised to convert it
                                              to band interleaving before proceeding.
                                            </p>
                                            <p>Do you want to convert this file for a more fluid viewing (this may take up additional disk space) ?</p>
                                            """,
                                            buttons = Qt.QMessageBox.StandardButton.No)
                    default_yes_button = Qt.QPushButton(self.style().standardIcon(Qt.QStyle.StandardPixmap.SP_DialogYesButton), "Yes (in default location)")
                    save_yes_button = Qt.QPushButton(self.style().standardIcon(Qt.QStyle.StandardPixmap.SP_DialogYesButton), "Yes (ask me where to save it)")

                    msgBox.addButton(default_yes_button, Qt.QMessageBox.ButtonRole.AcceptRole)
                    msgBox.addButton(save_yes_button, Qt.QMessageBox.ButtonRole.AcceptRole)
                    msgBox.setTextFormat(Qt.Qt.TextFormat.RichText)
                    msgBox.exec()
                    clicked = msgBox.clickedButton()
                    if clicked is save_yes_button:
                        band_dataset_path, _ = Qt.QFileDialog.getSaveFileName(self, "Save Band Dataset", self.__default_open_path, "GeoTiff Dataset (*.tiff)")
                        band_dataset_path = pathlib.Path(band_dataset_path)
                    elif clicked is not default_yes_button:
                        # User wants things to be slow
                        self._state.band_number = 0
                        self._state.dataset = dataset
                        return
                    if not self._convert_dataset_interleave(path, band_dataset_path):
                        return

                self._state.dataset = Dataset(band_dataset_path)
                self._state.band_number = 0
            else:
                self._state.dataset = dataset
                self._state.band_number = 0
        else:
            with open(path, 'r') as project_file:
                self._state.init_from_dict(json.load(project_file))
            self._project_file_path = pathlib.Path(path)

    @action("Save project", shortcut = "Ctrl+S")
    def save_project(self):
        if self._project_file_path is None:
            self.save_project_as()
        else:
            logger.info("Saving project as %s", self._project_file_path)
            json_string = json.dumps(self._state.to_dict(), indent=3)
            with self._project_file_path.open('w') as project_file:
                project_file.write(json_string)
    @action("Save project as...", shortcut = "Ctrl+Shift+S")
    def save_project_as(self):
        path, filetype = Qt.QFileDialog.getSaveFileName(self, "Save Project", self.__default_open_path, "InsarViz Project (*.invz)")
        if path == '':
            return
        path = pathlib.Path(path).with_suffix('.invz')
        self._project_file_path = path
        self.save_project()

    @action("Quit", shortcut = "Ctrl+Q")
    def quit(self):
        super().close()

    def _update_dataset(self):
        self._update_title()
    def _update_title(self):
        title = "InsarViz"
        band_title = self._state.current_band().band_title()
        if band_title is not None:
            title = f"{title} - {band_title}"
        self.setWindowTitle(title)

    def closeEvent(self, event):
        self._minimap_widget.close()
        self._dataset_widget.close()
        event.accept()

    @Qt.Slot(str) # type: ignore
    def _on_state_change(self, field):
        if field in ("dataset", "band_number"):
            self._update_title()



from .__prelude__ import Qt
import os

class PureShader:
    def __init__(self, shader_type, shader_source):
        self.shader_type = shader_type
        self.shader_source = shader_source
    def compile(self, program: Qt.QOpenGLShaderProgram):
        program.addShaderFromSourceCode(self.shader_type, self.shader_source)

class Shader(PureShader):
    def __init__(self, shader_type: Qt.QOpenGLShader.ShaderTypeBit, shader_path: str):
        with open(os.path.join(os.path.dirname(__file__),shader_path)) as shader_file:
            shader_source = shader_file.read()
        super().__init__(shader_type, shader_source)



from typing import Any, Optional
import json

from .__prelude__ import Qt, logger
from .decorators import dynamic

class ObservableStruct(Qt.QObject):
    fieldChanged = Qt.Signal(str, Any)

    def __init__(self):
        super().__init__()
        self.__dict__["__dynamic_attributes__"] = { }

    def dynamic_attribute(self, attr):
        sig = self.__dynamic_attributes__.get(attr, None)
        if sig is None:
            prop = getattr(self.__class__, attr, None)
            if isinstance(prop, dynamic.variable):
                sig = prop.register(self)
                sig.value_changed.connect(lambda old: self.fieldChanged.emit(attr, old))
                self.__dynamic_attributes__[attr] = sig
        return sig

    def __setattr__(self, name: str, value):
        sig = self.dynamic_attribute(name)
        if sig is None:
            old_val = getattr(self, name, None)
            super().__setattr__(name, value)
            self.fieldChanged.emit(name, old_val)
        else:
            sig.value = value

    @classmethod
    def from_dict(cls, dct: dict, /, **kwargs) -> Optional[Any]:
        ret = cls(**kwargs)
        ret.init_from_dict(dct)
        return ret
    def init_from_dict(self, __dct__):
        pass
    def to_dict(self) -> Optional[dict]:
        return None

    __mime_type__ = None
    @classmethod
    def can_from_mime(cls, mimeData, **kwargs):
        __kwargs__ = kwargs
        if cls.__mime_type__ is None:
            return False
        return mimeData.hasFormat(cls.__mime_type__)
    @classmethod
    def from_mime(cls, mimeData: Qt.QMimeData, **kwargs):
        if cls.__mime_type__ is None:
            return None
        bytes_data = mimeData.data(cls.__mime_type__).data()
        return cls.from_dict(json.loads(bytearray(bytes_data)), **kwargs)
    def to_mime(self) -> Optional[Qt.QMimeData]:
        if self.__mime_type__ is None:
            return None
        ret = Qt.QMimeData()
        dct = json.dumps(self.to_dict())
        ret.setData(self.__mime_type__, bytes(dct, "utf-8"))
        return ret



import numpy as np

from .__prelude__ import ObservableStruct, dynamic

class CurveParam(ObservableStruct):
    name       = dynamic.variable()
    min_bound  = dynamic.variable()
    max_bound  = dynamic.variable()
    initial    = dynamic.variable()

    def __init__(self, name, min_bound = -np.inf, max_bound = np.inf, initial = 0.0):
        super().__init__()
        self.name = name
        self.min_bound = min_bound
        self.max_bound = max_bound
        self.initial = initial

    @classmethod
    def from_dict(cls, dct, /, **kwargs):
        return CurveParam(dct['name'], dct['min_bound'], dct['max_bound'], dct['initial'])
    def init_from_dict(self, dct):
        self.name = dct["name"]
        self.min_bound = dct["min_bound"]
        self.max_bound = dct["max_bound"]
        self.initial = dct["initial"]
    def to_dict(self):
        return {
            'name': self.name,
            'min_bound': self.min_bound,
            'max_bound': self.max_bound,
            'initial': self.initial
        }



import ctypes
from OpenGL import GL
import numpy as np
from shiboken6 import VoidPtr

from ..__prelude__ import *

from ..observable    import (
    dynamic,
    ObservableStruct, ObservableList, SELF, EACH,
)
from ..SelectedBand  import SelectedBand
from ..MapPoint      import MapPoint
from ..MapProfile    import MapProfile
from ..ColorMap      import ColorMap, colormaps
from ..Dataset       import Dataset

def destroy_texture(tex):
    if tex is not None:
        tex.destroy()

class OpenGLPainter:
    def __init__(self, width, height):
        self._device = Qt.QOpenGLPaintDevice(width, height)
        self._painter = Qt.QPainter(self._device)
        self._painter.translate(0.0, height)
        self._painter.scale(1.0,-1.0)

    def __enter__(self):
        return self._painter
    def __exit__(self, *__args__):
        self._painter.end()



from .__prelude__ import Qt

from .Schema import Schema

class BoolSchema(Schema):
    class BoolNode(Schema.Node):
        def __init__(self, item_model, schema, model):
            super().__init__(item_model, schema)
            self._init_model(model)

        def flags(self):
            return Qt.Qt.ItemFlag.ItemIsUserCheckable | super().flags()
        def data(self, role):
            if role == Qt.Qt.ItemDataRole.DisplayRole:
                return self._schema._name
            if role == Qt.Qt.ItemDataRole.CheckStateRole:
                return Qt.Qt.CheckState.Checked if self.model else Qt.Qt.CheckState.Unchecked
        def setData(self, value, role):
            if role == Qt.Qt.ItemDataRole.CheckStateRole:
                self.requestModelChange.emit(value == Qt.Qt.CheckState.Checked.value)
                return True
            return False

        def _init_model(self, model):
            super()._init_model(model)
            self.dataChanged.emit([Qt.Qt.ItemDataRole.CheckStateRole])

    def __init__(self, name):
        super().__init__()
        self._name = name

    def make_node(self, item_model, model):
        return BoolSchema.BoolNode(item_model, self, model)



from typing import TypeVar
from .__prelude__ import SelectedBand, Matrix, Point, Qt, dynamic
from .GLDrawing import GLDrawing, InitializedDrawing

GenericScene = TypeVar("GenericScene", bound="Scene", covariant=True)

class SceneDrawing[GenericScene](InitializedDrawing[GenericScene]):
    class IdleMode:
        def __init__(self, initialized_scene):
            self.__initialized_scene = initialized_scene

        @property
        def drawing(self):
            return self.__initialized_scene.drawing
        @property
        def initialized_drawing(self):
            return self.__initialized_scene

        def default_transition(self):
            return self

        def on_mouse_enter(self, clip_position: tuple[float, float]):
            return self.default_transition()
        def on_mouse_move(self, clip_position: tuple[float,float]):
            return self.default_transition()
        def on_mouse_leave(self):
            return self.default_transition()

        def on_mouse_button_pressed(self, position: tuple[float, float], button: Qt.Qt.MouseButton):
            return self.default_transition()
        def on_mouse_button_released(self, position: tuple[float, float], button: Qt.Qt.MouseButton):
            return self.default_transition()

        def on_mouse_wheel(self, clip_position: tuple[float, float], angle_x: float, angle_y: float):
            return self.default_transition()

        def on_key_pressed(self, key: int):
            return self.default_transition()
        def on_key_released(self, key: int):
            return self.default_transition()

    def __init__(self, context: Qt.QOpenGLContext, scene: GenericScene):
        super().__init__(context, scene)

    def idle_mode(self):
        return self.IdleMode(self)

    @property
    def world_to_clip(self):
        return Matrix.identity(4)

    def clip_to_model(self, clip_x, clip_y):
        clip_to_model = Matrix.product(self.world_to_clip, self.drawing.model_to_world).inverse()

        v1_model = clip_to_model.transform_point((clip_x, clip_y, -0.5))
        v2_model = clip_to_model.transform_point((clip_x, clip_y, 0.5))
        ray_direction = v2_model - v1_model

        zMul = -v2_model[2]/ray_direction[2]
        return (v2_model[0]+ray_direction[0]*zMul, v2_model[1]+ray_direction[1]*zMul)

class Scene(GLDrawing):
    model_to_world  = dynamic.variable(Matrix.identity(4))

    def initializeDrawing(self, context, /):
        return SceneDrawing(context, self)



from .__prelude__ import Qt
from .WidgetTree import Container, Leaf

class RasterLayerDialog(Qt.QDialog):
    class ChildrenOfInterest:
        def __init__(self):
            self.band_chooser: Qt.QComboBox
            self.ok_button: Qt.QPushButton

    def __init__(self, dataset, parent = None):
        super().__init__(parent)

        self._dataset = dataset
        self.widgets = self.ChildrenOfInterest()
        wtree = Container(Qt.QVBoxLayout, (), [
            Leaf(Qt.QLabel, ("Choose band", )),
            Leaf(Qt.QComboBox, (), id="band_chooser"),
            Leaf(Qt.QPushButton, ("OK",), id="ok_button")
        ])
        wtree.create_in(self, self.widgets)
        self.setSizePolicy(Qt.QSizePolicy.Policy.Fixed, Qt.QSizePolicy.Policy.Fixed)

        self.setWindowTitle(f"Open dataset : {dataset.file}")
        self.setModal(True)
        for i, description in enumerate(dataset.descriptions):
            self.widgets.band_chooser.addItem(f"{i}: {description}", i)

        self.widgets.ok_button.clicked.connect(self.accept)

        self.setFixedSize(self.sizeHint())

    @property
    def band_number(self):
        return self.widgets.band_chooser.currentData()



import numpy as np

from .__prelude__ import (
    Qt, PureShader, Matrix, ColorMap, colormaps,
    dynamic, SELF,
    Dataset, destroy_texture
)
from .Layer import Layer, InitializedLayer

RASTER_SHADER = """
#version 330

uniform sampler2D {image};
uniform mat3 {geo_to_image};
uniform float {opacity};

float {geo_height}(vec3 geo_coords) {{
   return 0.0;
}}

vec4 {geo_color}(vec3 geo_coords, inout float up_opacity) {{
   vec3 image_coords = {geo_to_image}*geo_coords;
   vec4 value = texture({image}, image_coords.xy / image_coords.z);
   return vec4(value.xyz, {opacity}*up_opacity*value.w);
}}
"""

class RasterRGBLayer(Layer):
    opacity       = dynamic.variable(1.0)
    is_enabled    = dynamic.variable(True)
    image         = dynamic.readonly()
    geo_to_image  = dynamic.readonly()
    shaders       = dynamic.readonly()
    dataset       = dynamic.readonly()

    def __init__(self, dataset):
        super().__init__()
        self._dataset = dataset
        w,h = dataset.size
        first_band = dataset.read(1)
        self._image = np.reshape(
            np.stack([
                first_band,
                dataset.read(2),
                dataset.read(3),
                np.ones_like(first_band) * 255,
            ], axis=-1).astype(np.float32),
            (h,w,4)
        ) / 255.0
        self._geo_to_image = Matrix.product(
            dataset.pixel_to_crs,
            Matrix.scale((w,h,1))
        ).inverse()

        self._shaders = [
            RASTER_SHADER.format(
                **self.local_idents("image", "geo_to_image", "geo_color", "geo_height", "opacity")
            )
        ]
        self.dynamic_attribute("_redraw").value_changed.connect(lambda: self._redraw())

    @dynamic.method(SELF.opacity, SELF.is_enabled)
    def _redraw(self):
        self.renderChanged.emit()

    @property
    def description(self):
        return f"File: .../{self.dataset.file.name}"

    def GL_initialize(self, context):
        return InitializedRasterRGBLayer(context, self)

    __mime_type__ = "x-application/insarviz/RasterRGBLayer"
    @staticmethod
    def from_dict(dct, **__kwargs__):
        ret = RasterRGBLayer(Dataset(dct["dataset"]))
        ret.opacity = dct['opacity']
        ret.is_enabled = dct.get('is_enabled', True)
        return ret
    def to_dict(self):
        return {
            "opacity": self.opacity,
            "is_enabled": self.is_enabled,
            "dataset": str(self.dataset.file),
        }

class InitializedRasterRGBLayer(InitializedLayer):
    def __init__(self, context, layer):
        super().__init__(context, layer)

    @dynamic.memo(SELF.layer.image,
                     destroy = destroy_texture)
    def GL_texture(self):
        texture = Qt.QOpenGLTexture(Qt.QOpenGLTexture.Target.Target2D)
        img = self.layer.image
        h, w, _ = img.shape
        texture.setSize(w,h)
        texture.setWrapMode(Qt.QOpenGLTexture.WrapMode.ClampToBorder)
        texture.setBorderColor(Qt.QColor(0,0,0,0))
        texture.setFormat(Qt.QOpenGLTexture.TextureFormat.RGBA32F)
        texture.allocateStorage(Qt.QOpenGLTexture.PixelFormat.RGBA, Qt.QOpenGLTexture.PixelType.Float32)
        texture.setData(Qt.QOpenGLTexture.PixelFormat.RGBA, Qt.QOpenGLTexture.PixelType.Float32, img.data)
        texture.create()
        return texture

    @dynamic.memo(SELF.GL_texture, SELF.layer.opacity, SELF.layer.is_enabled)
    def GL_setup(self):
        image_texture = self.GL_texture
        def doit(program):
            uni = self.local_uniforms(program.uniforms)
            uni.image           = image_texture
            uni.opacity         = self.layer.opacity if self.layer.is_enabled else 0.0
            uni.geo_to_image    = self.layer.geo_to_image
        return doit



import rasterio
import rasterio.enums
import datetime
import numpy as np
import numpy.ma as npma
import pathlib
from threading import Lock

from .__prelude__ import Matrix, logger, ComputedValue
from .observable import ObservableStruct, dynamic, SELF

class AbstractDataset(ObservableStruct):
    def read(self, band, masked=False, **kwargs):
        arr = np.reshape(np.array([0], dtype=np.float32), (1,1))
        if masked:
            return npma.array(arr, mask = np.array([[False]]))
        else:
            return arr

    @property
    def file(self):
        return '<none>'
    @property
    def band_dates(self):
        return np.array([0], dtype=int)
    @property
    def count(self):
        return 1
    @property
    def width(self):
        return 1
    @property
    def height(self):
        return 1
    @property
    def band_timestamps(self):
        return np.array([0], dtype=float)
    @property
    def has_band_dates(self):
        return False
    @property
    def is_georeferenced(self):
        return False
    @property
    def size(self):
        return self.width, self.height

    @dynamic.memo()
    def value_units(self):
        return None

    @property
    def descriptions(self):
        return ["Dummy band"]

    @property
    def full_histograms(self):
        return ComputedValue(None)

    def nearest_band_to_timestamp(self, timestamp):
        after = np.searchsorted(self.band_timestamps, timestamp)
        left_i = after-1 if after > 0 else after
        right_i = after if after < self.count else after-1
        if abs(timestamp - self.band_timestamps[left_i]) < abs(timestamp - self.band_timestamps[right_i]):
            return left_i
        else:
            return right_i

    @property
    def nodatavals(self):
        return np.array([np.nan])
    @dynamic.memo()
    def pixel_to_crs(self):
        return Matrix.identity(3)
    @dynamic.memo(SELF.pixel_to_crs)
    def image_to_crs(self):
        w, h = self.width, self.height
        return Matrix.product(
            self.pixel_to_crs,
            Matrix.scale((w, h, 1))
        )

class Dataset(AbstractDataset):
    Interleaving = rasterio.enums.Interleaving

    def __init__(self, path):
        super().__init__()
        logger.info('Opening dataset at %s', path)
        self._dataset = rasterio.open(path)
        # RasterIO crashes when performing reads from multiple threads
        # on the same dataset, so we lock the reads
        self._dataset_lock = Lock()
        self._file = pathlib.Path(path)
        self._pixel_to_crs = Matrix(
            np.reshape(self._dataset.transform, (3,3)),
            np.reshape(~self._dataset.transform, (3,3))
        )
        datestamps = [self._get_date_and_stamps(i) for i in range(self._dataset.count)]
        self._band_dates = [date for _, date, _ in datestamps]
        self._band_timestamps = [stamp for _, _, stamp in datestamps]
        self._has_band_dates = any([has_date for has_date, _, _ in datestamps])
        logger.debug('Done opening dataset at %s', path)

    @dynamic.memo()
    def full_histograms(self):
        ret = ComputedValue(None)
        ret.recompute(self._compute_histograms)
        return ret

    def _get_date_and_stamps(self, i):
        tag_date = self._dataset.tags(i+1).get("DATE", None)
        if tag_date is not None:
            date = datetime.datetime.strptime(str(tag_date), "%Y%m%d")
            return True, date, date.timestamp()
        try:
            date = datetime.datetime.strptime(self._dataset.descriptions[i][-8:], "%Y%m%d")
            return True, date, date.timestamp()
        except (TypeError, ValueError):
            return False, i+1, float(i+1)

    def _compute_histograms(self):
        progress = 0.0
        progress_increment = 1.0/self.count

        sample_size = 1000
        samples = [None for _ in range(self.count)]
        fmt = "Full histograms : %p%"
        for band_number in range(self.count):
            band = self.read(band_number+1, masked=True)
            indices_x = np.random.randint(self.width,  size=sample_size)
            indices_y = np.random.randint(self.height, size=sample_size)
            indices_mask = ~band.mask[indices_y, indices_x]
            indices_x = indices_x[indices_mask]
            indices_y = indices_y[indices_mask]
            samples[band_number] = band.data[indices_y, indices_x]
            progress += progress_increment
            yield fmt, progress

        all_samples = np.concatenate(samples)
        hist, bins = np.histogram(all_samples, bins='fd', density=True)
        return hist, bins

    def read(self, band, masked=False, **kwargs):
        with self._dataset_lock:
            raw_img = self._dataset.read(band, masked=masked, **kwargs)
        if masked and len(raw_img.mask.shape) != 2:
            nodata = self._dataset.nodata
            if nodata is None:
                logger.warn(f"Nodata mask not found in {self.file}. Using dark magic to create one")
                # Some datasets use 9999.0 or -9999.0 as indicators of "nodata", without telling us
                mask = np.logical_or(np.logical_or(raw_img.data == 9999.0, raw_img.data == -9999.0),
                                     np.logical_or(raw_img.data == 0.0, np.isnan(raw_img.data)))
                raw_img = npma.array(raw_img.data, mask = mask)
            else:
                if np.isnan(nodata):
                    raw_img = npma.array(raw_img.data, mask = np.isnan(raw_img.data))
                else:
                    raw_img = npma.masked_values(raw_img.data, nodata)

        return raw_img

    @dynamic.memo()
    def metadata(self):
        metafilename = self.file.with_suffix('.meta')
        metadata = {}
        try:
            with open(metafilename, encoding="utf-8") as f:
                for line in f:
                    try:
                        (key, val) = line.split(sep=': ', maxsplit=1)
                        metadata[key] = val.strip()
                    except ValueError:
                        pass
        except FileNotFoundError:
            logger.warn("No metadata file found")

        return metadata

    @dynamic.memo()
    def value_units(self):
        tags = self._dataset.tags()
        if "VALUE_UNITS" in tags:
            units = tags["VALUE_UNIT"]
            logger.info("Found value units '%s' in dataset tags", units)
            return units
        metadata = self.metadata
        if "Value_unit" in metadata:
            units = metadata["Value_unit"]
            logger.info("Found value units '%s' in metadata file", units)
            return units
        return None

    @property
    def interleaving(self):
        return self._dataset.interleaving

    @property
    def is_georeferenced(self):
        return self._dataset.crs is not None

    @property
    def file(self):
        return self._file

    @property
    def band_dates(self):
        return self._band_dates
    @property
    def has_band_dates(self):
        return self._has_band_dates
    @property
    def band_timestamps(self):
        return self._band_timestamps
    @property
    def descriptions(self):
        return self._dataset.descriptions

    @property
    def count(self):
        return self._dataset.count

    @property
    def nodatavals(self):
        return self._dataset.nodatavals
    @property
    def width(self):
        return self._dataset.width
    @property
    def height(self):
        return self._dataset.height

    @property
    def pixel_to_crs(self):
        return self._pixel_to_crs



from .__prelude__ import Qt
from .WidgetTree import Leaf, Container

class HelpWidget(Qt.QWidget):
    class ChildrenOfInterest:
        pass

    def __init__(self):
        super().__init__()

        self.widgets = self.ChildrenOfInterest()
        def mkGroupBox(title):
            ret = Qt.QGroupBox(title)
            ret.setFlat(True)
            return ret
        def separator():
            ret = Qt.QWidget()
            ret.setFixedHeight(1)
            ret.setSizePolicy(Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Fixed)
            ret.setStyleSheet("background-color: #c0c0c0;")
            return ret

        wtree = Container(Qt.QHBoxLayout, (), [
            Container(Qt.QVBoxLayout, (), [
                Leaf(Qt.QLabel, ("F1 : Show/hide this help", )),
                Leaf(separator, ()),
                Leaf(Qt.QLabel, ("Ctrl+S : Save project", )),
                Leaf(Qt.QLabel, ("Ctrl+Shift+S : Save project as...", )),
                Leaf(Qt.QLabel, ("Ctrl+O : Open dataset", )),
                Leaf(separator, ()),
                Leaf(Qt.QLabel, ("Ctrl+P : Show/hide plots", )),
                Leaf(Qt.QLabel, ("Ctrl+M : Show/hide minimap", )),
                (Leaf(Qt.QWidget, ()), {"stretch":1})
            ], widget_class=lambda: mkGroupBox("Global interaction")),
            Container(Qt.QVBoxLayout, (), [
                Leaf(Qt.QLabel, ("Mouse wheel : zoom in/out", )),
                Leaf(Qt.QLabel, ("Ctrl+Mouse wheel : rotate plane", )),
                Leaf(Qt.QLabel, ("Shift+Mouse wheel : tilt plane", )),
                Leaf(Qt.QLabel, ("Ctrl+Shift+Mouse wheel : Increase/decrease height LOS scaling", )),
                Leaf(separator, ()),
                Leaf(Qt.QLabel, ("Mouse drag : drag plane", )),
                Leaf(Qt.QLabel, ("Ctrl+Mouse drag : create point with radius", )),
                Leaf(Qt.QLabel, ("Shift+Mouse drag : create profile (hold Shift for multiple end points)", )),
                Leaf(separator, ()),
                Leaf(Qt.QLabel, ("Ctrl+A : Autorange current band", )),
                Leaf(Qt.QLabel, ("Ctrl+Shift+A : Autorange all bands", )),
            ], widget_class=lambda: mkGroupBox("Map interaction")),
        ])
        wtree.create_in(self, self.widgets)



from .__prelude__ import (
    Qt,
    MapPoint, PointCurve, LinearCurve, SigmoidCurve, PeriodicCurve, HeavisideCurve,
    StructRoles, StructSchema, MaybeStructSchema, FloatSliderSchema, DateSchema,
    BoolSchema
)

class PlotPointRoles(StructRoles):
    @staticmethod
    def flags():
        return Qt.Qt.ItemFlag.ItemIsEnabled | Qt.Qt.ItemFlag.ItemIsEditable | Qt.Qt.ItemFlag.ItemIsUserCheckable
    @staticmethod
    def field_roles(field):
        if field == "color":
            return [Qt.Qt.ItemDataRole.DecorationRole]
        if field == "name":
            return [Qt.Qt.ItemDataRole.DisplayRole]
        if field == "show_in_map":
            return [Qt.Qt.ItemDataRole.CheckStateRole]
        return None
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return model.name
        if role == Qt.Qt.ItemDataRole.DecorationRole:
            return model.color
        if role == Qt.Qt.ItemDataRole.CheckStateRole:
            return Qt.Qt.CheckState.Checked if model.show_in_plot else Qt.Qt.CheckState.Unchecked
    @staticmethod
    def set_role(model, role, value):
        if role == Qt.Qt.ItemDataRole.EditRole:
            model.name = value
            return True
        if role == Qt.Qt.ItemDataRole.CheckStateRole:
            model.show_in_plot = (value == Qt.Qt.CheckState.Checked.value)
            return True
        return False
class PointCurveRoles(StructRoles):
    @staticmethod
    def flags():
        return Qt.Qt.ItemFlag.ItemIsEnabled
    @staticmethod
    def field_roles(field):
        if field == "is_empty":
            return [Qt.Qt.ItemDataRole.CheckStateRole]
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Fit Curve"
        if role == Qt.Qt.ItemDataRole.CheckStateRole:
            return Qt.Qt.CheckState.Checked if not model.is_empty else Qt.Qt.CheckState.Unchecked
class SigmoidCurveRoles(StructRoles):
    @staticmethod
    def flags():
        return Qt.Qt.ItemFlag.ItemIsEnabled
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Sigmoid Curve"
class LinearCurveRoles(StructRoles):
    @staticmethod
    def flags():
        return Qt.Qt.ItemFlag.ItemIsEnabled
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Linear Curve"
class SeasonalCurveRoles(StructRoles):
    @staticmethod
    def flags():
        return Qt.Qt.ItemFlag.ItemIsEnabled
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Seasonal Curve"
class SemiSeasonalCurveRoles(StructRoles):
    @staticmethod
    def flags():
        return Qt.Qt.ItemFlag.ItemIsEnabled
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Semi-Seasonal Curve"
class HeavisideCurveRoles(StructRoles):
    @staticmethod
    def flags():
        return Qt.Qt.ItemFlag.ItemIsEnabled
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Post-seismic Curve"

point_curve_schema = StructSchema(
    PointCurve, PointCurveRoles,
    ("linear_curve",
     MaybeStructSchema(
         LinearCurve, LinearCurveRoles)),
    ("sigmoid_curve",
     MaybeStructSchema(
         SigmoidCurve, SigmoidCurveRoles,
         ("initial_step", FloatSliderSchema("Step position", f_min = -1.0, f_max = 1.0)))),
    ("seasonal_curve",
     MaybeStructSchema(
         PeriodicCurve, SeasonalCurveRoles)),
    ("semi_seasonal_curve",
     MaybeStructSchema(
         PeriodicCurve, SemiSeasonalCurveRoles)),
    ("postseismic_curve",
     MaybeStructSchema(
         HeavisideCurve, HeavisideCurveRoles)),
    ("seism_timestamp", DateSchema("Seism date"))
)
point_schema = StructSchema(
    MapPoint, PlotPointRoles,
    ("show_variance", BoolSchema("Standard deviation")),
    ("point_curve",point_curve_schema))
focus_point_schema = StructSchema(
    MapPoint, PlotPointRoles,
    ("show_variance", BoolSchema("Standard deviation")),
    ("point_curve",point_curve_schema),
    ("focus", FloatSliderSchema("Position"))
)



from typing import Optional

from .__prelude__ import Qt

class IconDockStyle(Qt.QProxyStyle):

    # adapted from https://stackoverflow.com/a/3482795

    def __init__(self, icon: Qt.QIcon, style: Optional[Qt.QStyle] = None):
        super().__init__(style)
        self.icon = icon

    def drawControl(self, element: Qt.QStyle.ControlElement, option: Qt.QStyleOption, painter: Qt.QPainter,
                    widget: Optional[Qt.QWidget] = None) -> None:
        if element == Qt.QStyle.ControlElement.CE_DockWidgetTitle:
            # width of the icon
            width: int = self.pixelMetric(Qt.QStyle.PixelMetric.PM_TabBarIconSize)
            # margin of title from frame
            margin: int = self.baseStyle().pixelMetric(Qt.QStyle.PixelMetric.PM_DockWidgetTitleMargin)
            icon_point = Qt.QPoint(margin + option.rect.left(),
                                   margin + option.rect.center().y() - width//2)
            painter.drawPixmap(icon_point, self.icon.pixmap(width, width))
            option.rect = option.rect.adjusted(width, 0, 0, 0)
        self.baseStyle().drawControl(element, option, painter, widget)


def add_icon_to_tab(window: Qt.QMainWindow, tab_name: str, icon: Qt.QIcon):
    # adapted from https://stackoverflow.com/a/46623219
    def f(_):
        try:
            for tabbar in window.findChildren((Qt.QTabBar)):
                for i in range(tabbar.count()):
                    if tabbar.tabText(i) == tab_name:
                        tabbar.setTabIcon(i, icon)
        except RuntimeError:
            # window has been deleted
            pass
    return f


class IconDockWidget(Qt.QDockWidget):

    def __init__(self, name: str, parent: Qt.QWidget, icon: Qt.QIcon):
        assert isinstance(parent, Qt.QMainWindow)
        super().__init__(name, parent)
        self.setStyle(IconDockStyle(icon))
        self.__callback = add_icon_to_tab(parent, name, icon)
        self.visibilityChanged.connect(self.__callback)
        self.topLevelChanged.connect(self.__callback)
        self.dockLocationChanged.connect(self.__callback)
        self.hide()



from typing import Any, Optional, override

from .Schema import Schema
from .__prelude__ import Qt

class IntSchema(Schema):
    class IntNode(Schema.Node):
        def __init__(self, item_model: Qt.QAbstractItemModel, schema: "IntSchema", model: int, name: str):
            super().__init__(item_model, schema)
            self.name: str = name
            self._editor: Optional[Qt.QSpinBox] = None
            self.set_model(model)

        def createEditor(self, parent: Qt.QWidget) -> Optional[Qt.QWidget]:
            if self._editor is None:
                self._editor = Qt.QSpinBox()
                self._editor.setFrame(False)
                self._editor.setMinimum(0)
                self._editor.setMaximum(10000)
                self._editor.valueChanged.connect(lambda value: self.requestModelChange.emit(value))
            self._editor.setPrefix(f"{self.name} : ")
            self._editor.setParent(parent)
            return self._editor
        def destroyEditor(self) -> bool:
            return False
        def editor(self) -> Optional[Qt.QWidget]:
            return self._editor
        def setEditorData(self, editor):
            if self.model is not None:
                editor.setValue(self.model)
            return True

        def flags(self):
            return Qt.Qt.ItemFlag.ItemIsEditable | super().flags()
        def set_model(self, model: int):
            super().set_model(model)
            if self._editor is not None:
                self._editor.setValue(model)
            self.dataChanged.emit([])

    def __init__(self, name):
        self.name = name

    def make_node(self, item_model: Qt.QAbstractItemModel, model: Any) -> "Schema.Node":
        return self.IntNode(item_model, self, model, self.name)

class IntSliderSchema(Schema):
    class IntSliderWidget(Qt.QGroupBox):
        valueChanged = Qt.Signal(int)

        def __init__(self, name):
            super().__init__(name)
            layout = Qt.QHBoxLayout(self)
            self._layout = layout
            self._label = Qt.QLabel(name)
            slider = Qt.QSlider(orientation=Qt.Qt.Orientation.Horizontal)
            slider.valueChanged.connect(self.valueChanged.emit)
            layout.addWidget(slider, stretch=1)
            self._slider : QSlider = slider
            layout.setContentsMargins(5,2,5,2)
            self.setLayout(layout)

        def value(self):
            return self._slider.value()
        def setValue(self, value):
            return self._slider.setValue(value)

    class IntSliderNode(Schema.Node):
        def __init__(self, item_model, schema, model, name):
            super().__init__(item_model, schema)
            self.name = name
            self._create_editor()
            self.set_model(model)

        def _create_editor(self):
            self._editor = IntSliderSchema.IntSliderWidget(self.name)
            self._editor.valueChanged.connect(self.requestModelChange.emit)

        def createEditor(self, parent):
            self._editor.setParent(parent)
            return self._editor
        def destroyEditor(self):
            return False
        def editor(self):
            return self._editor
        def setModelData(self, __editor__):
            return True
        def setEditorData(self, editor):
            editor.setValue(self.model)
            return True

        def flags(self):
            return Qt.Qt.ItemFlag.ItemIsEditable | super().flags()
        def set_model(self, model):
            super().set_model(model)
            self.dataChanged.emit([])

        # def data(self, role):
        #     if role == Qt.ItemDataRole.DisplayRole:
        #         return self.model
        #     return None

    def __init__(self, name):
        self.name = name

    def make_node(self, item_model, model):
        return self.IntSliderNode(item_model, self, model, self.name)



from .__prelude__ import Qt, Matrix

class WidgetTree:
    def __init__(self, id = None):
        self._id = id
    def create(self, widget_ids):
        raise NotImplementedError
    def register(self, widget_ids, w):
        if self._id is not None and not hasattr(widget_ids, self._id):
            setattr(widget_ids, self._id, w)
class Container(WidgetTree):
    def __init__(self, layout_class, layout_args, children, widget_class = Qt.QWidget, id = None):
        super().__init__(id)
        self._widget_class = widget_class
        self._layout_class = layout_class
        self._layout_args = layout_args
        self._children = children

    def create(self, widget_ids):
        ret = self._widget_class()
        self.create_in(ret, widget_ids)
        return ret
    def create_in(self, w, widget_ids):
        layout = self._layout_class(*self._layout_args)
        w.setLayout(layout)
        for child in self._children:
            if isinstance(child, tuple):
                layout.addWidget(child[0].create(widget_ids), **child[1])
            else:
                layout.addWidget(child.create(widget_ids))
        self.register(widget_ids, w)
class Leaf(WidgetTree):
    def __init__(self, widget_class, widget_args, widget_kwargs = {}, id = None):
        self._widget_class = widget_class
        self._widget_args = widget_args
        self._widget_kwargs = widget_kwargs
        self._id = id
    def create(self, widget_ids):
        ret = self._widget_class(*self._widget_args, **self._widget_kwargs)
        self.register(widget_ids, ret)
        return ret



import numpy as np

from .__prelude__ import dynamic
from .CurveParam import CurveParam
from .Curve import Curve

class HeavisideCurve(Curve):
    step = dynamic.variable()

    def __init__(self, curve_after):
        super().__init__()
        self.step = 0.0
        self.curve_after = curve_after

    __mime_type__ = "application/x-insarviz/HeavisideCurve"

    def to_dict(self):
        return {
            "step": self.step,
            "curve_after": self.curve_after.to_dict()
        }
    def init_from_dict(self, dct):
        self.step = dct["step"]
        self.curve_after.init_from_dict(dct["curve_after"])

    @property
    def nparams(self):
        return self.curve_after.nparams
    @property
    def params(self):
        return self.curve_after.params

    def func(self, xs, *params):
        return np.where(xs <= self.step, 0.0, self.curve_after.func(xs, *params))
    def jacobian(self, xs, *params):
        return np.where(xs <= self.step, 0.0, self.curve_after.jacobian(xs, *params))
    def hessian_epsilon(self, xs, err, grad, *params):
        xs_filter = xs >= self.step
        return self.curve_after.hessian_epsilon(xs[xs_filter], err[xs_filter], grad, *params)



import numpy as np

from .CurveParam import CurveParam
from .Curve import Curve

class LogCurve(Curve):
    def __init__(self):
        super().__init__()
        self.offset = CurveParam("offset")
        self.offset.min_bound = 1.00001
        self.offset.initial = 2.0
        self.scale = CurveParam("scale")

    __mime_type__ = "application/x-insarviz/LogCurve"

    @property
    def nparams(self):
        return 2
    @property
    def params(self):
        return [self.scale, self.offset]

    def func(self, xs, scale, offset):
        return scale * np.log(xs + offset)
    def jacobian(self, xs, scale, offset):
        offset_xs = xs + offset
        return np.array([np.log(offset_xs), scale / offset_xs])
    def hessian_epsilon(self, xs, err, grad, scale, offset):
        offset_xs = xs + offset

        d_scale_scale    = 0
        d_scale_offset   = np.sum(err / offset_xs)
        d_offset_offset  = - scale * np.sum(err / (offset_xs * offset_xs))

        hessian = np.array([
            [ d_scale_scale, d_scale_offset ],
            [ d_scale_offset, d_offset_offset]
        ])

        return np.dot(hessian @ grad, grad)



from typing import Optional
from OpenGL import GL

from .__prelude__ import (
    SelectedBand, Shader, Qt, Matrix, Point, Bound, GLProgram,
    dynamic, SELF,
    destroy_texture,
)

from .Scene import Scene, SceneDrawing
from .MapScene import MapScene

class InitializedMinimapScene(SceneDrawing["MinimapScene"]):
    class IdleMode(SceneDrawing.IdleMode):
        def on_mouse_button_pressed(self, position, button):
            if button == Qt.Qt.MouseButton.LeftButton:
                return self.initialized_drawing.DraggingMode(self.initialized_drawing, position)
            return self

    class DraggingMode(SceneDrawing.IdleMode):
        def __init__(self, initialized_scene, start_position):
            super().__init__(initialized_scene)
            self.__start_position = start_position

        def on_mouse_move(self, position):
            xa,ya = self.initialized_drawing.clip_to_model(*self.__start_position)
            xb,yb = self.initialized_drawing.clip_to_model(*position)
            self.drawing.map_scene.center += Point(xa-xb, ya-yb)
            return self.initialized_drawing.DraggingMode(self.initialized_drawing, position)
        def on_mouse_button_released(self, position, button):
            return self.initialized_drawing.IdleMode(self.initialized_drawing)

    def __init__(self, context, scene):
        super().__init__(context, scene)
        self._program = GLProgram(self.drawing._shaders)
        self._image_mesh = self._program.create_square_mesh(GL.GL_TRIANGLES)
        self._rect_program = GLProgram(self.drawing._rect_shaders)
        self._loop_mesh = self._rect_program.create_square_mesh(GL.GL_LINE_LOOP)
        self.map_world_to_clip = Matrix.identity(4)

    @dynamic.memo(SELF.viewport_size)
    def world_to_clip(self):
        w, h = self.viewport_size
        hfactor = min(1.0, float(h)/float(w))
        vfactor = min(1.0, float(w)/float(h))

        return Matrix.product(
            Matrix.scale((hfactor, vfactor, 1.0, 1.0))
        )

    @dynamic.memo(SELF.drawing.colormap.createTexture,
                     destroy = destroy_texture)
    def GL_colormap_texture(self):
        return self.drawing.colormap.createTexture()

    def paintGL(self):
        gl = self.context.functions()
        gl.glEnable(GL.GL_BLEND)
        gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
        gl.glClearColor(0.5, 0.5, 0.5, 1.0)
        gl.glClear(int(GL.GL_COLOR_BUFFER_BIT))

        colormap_texture = self.GL_colormap_texture
        dem_texture = self.drawing.selected_band.texture
        ref_texture = self.drawing.selected_band.reference_texture
        with self._program as p:
            p.uniforms.image             = dem_texture
            p.uniforms.reference_image   = ref_texture
            p.uniforms.model_to_clip     = self.world_to_clip * self.drawing.model_to_world
            p.uniforms.colormap          = colormap_texture
            p.uniforms.value_to_color    = self.drawing._colormap.value_to_color
            p.uniforms.texture_to_image  = dem_texture.texture_to_image
            p.uniforms.model_to_texture  = Matrix.product(
                self.drawing.map_scene.model_to_texture,
                Matrix.translate((-1,-1)),
                Matrix.scale((2,2,1)),
            )
            self._image_mesh.draw()
        with self._rect_program as p:
            p.uniforms.clip_to_map_model = Matrix.product(
                self.map_world_to_clip,
                self.drawing._map_scene.model_to_world
            ).inverse()
            p.uniforms.model_to_clip = self.world_to_clip * self.drawing.model_to_world
            p.uniforms.rect_color = Qt.QColor("white")
            self._loop_mesh.draw()

    def freeGL(self):
        self.dynamic_attribute("GL_colormap_texture").clear()

class MinimapScene(Scene):
    colormap = dynamic.readonly()
    selected_band = dynamic.readonly()
    map_scene = dynamic.readonly()

    def __init__(self, selected_band: SelectedBand, map_scene: MapScene, colormap):
        super().__init__()
        self._selected_band = selected_band
        self._map_scene = map_scene
        self._colormap = colormap
        self._shaders = [
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Vertex, "minimap/vertex.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment, "minimap/fragment.glsl")
        ]
        self._rect_shaders = [
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Vertex, "viewport_rect/vertex.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment, "viewport_rect/fragment.glsl")
        ]
        self.dynamic_attribute('_redraw').value_changed.connect(lambda: self._redraw())

    @dynamic.method(SELF.selected_band.image, SELF.selected_band.reference_image,
                    SELF.colormap.image, SELF.colormap.value_to_color,
                    SELF.map_scene.model_to_texture,
                    SELF.map_scene.model_to_world)
    def _redraw(self):
        self.renderChanged.emit()

    @dynamic.memo()
    def model_to_world(self):
        return Matrix.identity(4)
    @dynamic.memo()
    def model_to_texture(self):
        return Matrix.product(
            Matrix.translate((0.5,0.5)),
            Matrix.scale((0.5,0.5,1.0))
        )

    def initializeDrawing(self, context, /):
        return InitializedMinimapScene(context, self)



import numpy as np
from typing import Any, Optional, override
import pyqtgraph

from .__prelude__ import (Qt, colormaps, Leaf, Container)
from .Schema import Schema

def make_colormap(name, values, mode):
    formatted_values = np.rint(np.array(values)*255).astype(int)
    ticks = np.linspace(0., 1., len(formatted_values))
    return pyqtgraph.ColorMap(ticks, formatted_values, name=name, mapping=mode)
pg_colormaps = { name: make_colormap(name, values, pyqtgraph.ColorMap.CLIP) for name, values in colormaps.items() }
pg_colormaps.update({ 'cyclic '+name: make_colormap('cyclic '+name, values, pyqtgraph.ColorMap.REPEAT) for name, values in colormaps.items() })

class ColormapEditor(pyqtgraph.HistogramLUTWidget):
    def __init__(self, colormap):
        super().__init__()
        self._colormap = colormap
        self.setBackground('w')
        self.setForegroundBrush(Qt.Qt.GlobalColor.transparent)

        self._hist_graph: pyqtgraph.BarGraphItem = pyqtgraph.BarGraphItem(x0=[],x1=[],y0=[],y1=[])
        self._hist_graph.setRotation(90)
        self._hist_graph.setOpts(brush = Qt.QColor(220,20,20,100), pen=None)
        hist, bins = colormap.image_histogram
        self._hist_graph.setOpts(x0 = bins[:-1], x1 = bins[1:], y0 = 0, y1 = hist)
        self.vb.addItem(self._hist_graph)

        self._init_colormap_menu()
        self._set_histogram_levels()
        self.setLevels(self._colormap.xzero, self._colormap.xone)
        self.setHistogramRange(self._colormap.xzero, self._colormap.xone, padding=0.1)
        self.sigLevelsChanged.connect(self._on_levels_changed)

    def _set_histogram_levels(self, set_levels = False):
        hist, bins = self._colormap.image_histogram
        hist_norm = hist/np.sum(hist)
        hist_cum = np.cumsum(hist_norm)
        lower, upper = np.array(bins)[np.searchsorted(hist_cum, [0.1, 0.9])]
        self.vb.setLimits(yMin=bins[0], yMax=bins[-1])
        self.setHistogramRange(lower,upper, padding=0.1)
        if set_levels:
            self.setLevels(min=lower, max=upper)

    @Qt.Slot(Any) #type: ignore
    def _on_levels_changed(self, item):
        xmin, xmax = item.getLevels()
        self._colormap.xzero = xmin
        self._colormap.xone = xmax

    def sizeHint(self):
        return Qt.QSize(200, 600)

    def auto_range(self):
        self._set_histogram_levels(True)

    def _init_colormap_menu(self):
        grad = self.gradient

        grad.menu = pyqtgraph.ColorMapMenu(
            userList=[
                pg_colormaps[name] for name in colormaps.keys()
            ] + [
                pg_colormaps['cyclic '+name] for name in colormaps.keys()
            ])
        grad.menu.removeAction(grad.menu.actions()[0]) # Remove the default "None" colormap
        grad.menu.sigColorMapTriggered.connect(grad.colorMapMenuClicked)
        grad.menu.sigColorMapTriggered.connect(self._on_colormap_selected)
        grad.colorMapMenuClicked(pg_colormaps[self._colormap.name])

    def _on_colormap_selected(self, colormap):
        self._colormap.name = colormap.name
        self._colormap.image = colormap.getLookupTable(nPts=1024)
        self._colormap.is_cyclic = colormap.mapping_mode == pyqtgraph.ColorMap.REPEAT

class ColorMapSchema(Schema):
    class ColorMapNode(Schema.Node):
        def __init__(self, item_model: Qt.QAbstractItemModel, schema: "ColorMapSchema", model):
            super().__init__(item_model, schema)
            model.fieldChanged.connect(self._on_colormap_changed)
            self.set_model(model)

        @Qt.Slot(str)
        def _on_colormap_changed(self, field):
            if field == "image":
                self.dataChanged.emit([Qt.Qt.ItemDataRole.DisplayRole])

        def flags(self):
            return Qt.Qt.ItemFlag.ItemIsEnabled | Qt.Qt.ItemFlag.ItemIsEditable | super().flags()
        def data(self, role):
            if role == Qt.Qt.ItemDataRole.SizeHintRole:
                return Qt.QSize(100, 20)
            if role == Qt.Qt.ItemDataRole.DisplayRole:
                return ""
            return None

        def createEditor(self, parent):
            self._dialog = Qt.QDialog(parent)
            self._dialog.setWindowFlags(self._dialog.windowFlags() | Qt.Qt.WindowStaysOnTopHint)
            self._dialog.setWindowTitle(f"Colormap : {self.model.name}")
            widgets = type('', (object,), {})()

            wtree = Container(Qt.QVBoxLayout, (), [
                (Leaf(ColormapEditor, (self.model, ), id="colormap_editor"), {"stretch":1}),
                Leaf(Qt.QPushButton, ("Autorange", ), id="autorange_button"),
            ])

            wtree.create_in(self._dialog, widgets)
            widgets.autorange_button.clicked.connect(lambda: widgets.colormap_editor.auto_range())

            self._dialog.show()
            return self._dialog
        def destroyEditor(self):
            self._dialog.close()
            self._dialog = None
            return False

        def paint(self, painter, option):
            rect = option.rect.adjusted(5,3,-5,-3)
            line = Qt.QImage((np.tile(self.model.interpolated(rect.width()), (rect.height(), 1))).astype(np.uint8),
                          rect.width(), rect.height(), rect.width()*3, Qt.QImage.Format.Format_RGB888)
            painter.drawImage(rect.x(), rect.y(), line)
            return True

    def __init__(self, name):
        self.name = name

    def make_node(self, item_model: Qt.QAbstractItemModel, model: Any) -> "Schema.Node":
        return self.ColorMapNode(item_model, self, model)



from pathlib import Path
import urllib.request
import urllib.error
import os
from OpenGL import GL
import numpy as np
import math

from .__prelude__ import (
    Qt, logger, insarviz_dirs, ObservableStruct, dynamic, SELF,
    GLProgram, Shader, Matrix, Bound,
)

class TileRequest(Qt.QRunnable):
    class Signals(Qt.QObject):
        tile_downloaded = Qt.Signal(str)

    def __init__(self, tile_path, tile_url, tile_file):
        super().__init__()
        self._tile_path = tile_path
        self._tile_file = tile_file
        self._tile_url = tile_url
        self._signals = TileRequest.Signals()

    @property
    def tile_downloaded(self):
        return self._signals.tile_downloaded

    def run(self):
        tile_file = self._tile_file
        tile_url = self._tile_url
        req = urllib.request.Request(tile_url, headers = {
            "User-Agent": "InsarViz/3.0 map view"
        })
        os.makedirs(tile_file.parent, exist_ok=True)
        try:
            image = urllib.request.urlopen(req).read()
            with tile_file.open(mode="wb") as f:
                f.write(image)
            self.tile_downloaded.emit(self._tile_path)
        except urllib.error.HTTPError:
            logger.warn("Got HTTP error when downloading tile %s", tile_url)

class TileCache(ObservableStruct):
    new_tile_available = Qt.Signal()
    provider = dynamic.variable()

    def __init__(self):
        super().__init__()
        self._tile_cache = { }
        self._tile_requests = { }
        self.provider = None
        self.dynamic_attribute("provider").value_changed.connect(self.__clear_image_cache)

    @dynamic.method(SELF.provider)
    def get_tile(self, x_tile, y_tile, zoom_level):
        tile_path = f"{zoom_level}/{x_tile}/{y_tile}"
        tile_count = 2**zoom_level
        if x_tile < 0 or x_tile >= tile_count:
            return None
        if y_tile < 0 or y_tile >= tile_count:
            return None

        if not tile_path in self._tile_cache:
            tile_file = Path(insarviz_dirs.user_cache_dir) / "tiles" / self.provider.name / f"{tile_path}.png"
            if not tile_path in self._tile_requests:
                if tile_file.exists():
                    self._tile_cache[tile_path] = Qt.QImage(tile_file)
                else:
                    req = TileRequest(tile_path, self.provider.build_url(z=zoom_level, x=x_tile, y=y_tile), tile_file)
                    req.tile_downloaded.connect(self._on_tile_downloaded)
                    self._tile_requests[tile_path] = req
                    logger.debug("Downloading tile %s", tile_path)
                    Qt.QThreadPool.globalInstance().start(req)
                    return None
            else:
                return None
        return self._tile_cache[tile_path]

    @Qt.Slot()
    def __clear_image_cache(self):
        self._tile_cache = { }
        self.new_tile_available.emit()

    def _on_tile_downloaded(self, tile_path):
        del self._tile_requests[tile_path]
        self.new_tile_available.emit()

    def GL_paint_background(self, glfunc, model_to_crs, clip_to_model, model_to_world, world_to_clip, tiles_per_screen):
        tile_program = GLProgram([
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Vertex,"tile/vertex.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment,"tile/fragment.glsl"),
        ])
        tile_mesh = tile_program.create_square_mesh(GL.GL_TRIANGLES, split=0, texture_to_model = Matrix.identity(3))

        def clip_to_crs(x_clip, y_clip):
            x_model, y_model = clip_to_model(x_clip, y_clip)
            return model_to_crs.transform_point((x_model, y_model))
        left_x_crs, left_y_crs = clip_to_crs(-1.0, 0.5)
        right_x_crs, right_y_crs = clip_to_crs(1.0, 0.5)

        d_crs = np.array([
            right_x_crs - left_x_crs,
            right_y_crs - left_y_crs
        ])

        crs_distance = np.sqrt(np.dot(d_crs, d_crs))
        zoom_level = int(- np.ceil(np.log2(crs_distance / (360.0 * tiles_per_screen))))


        topleft_x_crs, topleft_y_crs          = clip_to_crs(-1.0, 1.0)
        topright_x_crs, topright_y_crs        = clip_to_crs(1.0,  1.0)
        bottomleft_x_crs, bottomleft_y_crs    = clip_to_crs(-1.0, -1.0)
        bottomright_x_crs, bottomright_y_crs  = clip_to_crs(1.0,  -1.0)

        def crs_to_tile(x_crs, y_crs):
            n = 2**zoom_level
            x_tile = int(n * ((x_crs + 180.0) / 360.0))
            lat_rad = y_crs * np.pi / 180.0
            while lat_rad > np.pi:
                lat_rad -= 2*np.pi
            while lat_rad <= -np.pi:
                lat_rad += 2*np.pi
            y_tile = int(n * (1 - (np.log(math.tan(lat_rad) + 1/math.cos(lat_rad)) / np.pi)) / 2.0)
            return x_tile, y_tile

        x_start, x_end = (
            min(topleft_x_crs, topright_x_crs, bottomleft_x_crs, bottomright_x_crs),
            max(topleft_x_crs, topright_x_crs, bottomleft_x_crs, bottomright_x_crs)
        )
        y_start, y_end = (
            min(topleft_y_crs, topright_y_crs, bottomleft_y_crs, bottomright_y_crs),
            max(topleft_y_crs, topright_y_crs, bottomleft_y_crs, bottomright_y_crs)
        )
        x_left_tile, y_left_tile = crs_to_tile(x_start, y_start)
        x_right_tile, y_right_tile = crs_to_tile(x_end, y_end)
        x_left_tile, x_right_tile = min(x_left_tile, x_right_tile), max(x_left_tile, x_right_tile)
        y_left_tile, y_right_tile = min(y_left_tile, y_right_tile), max(y_left_tile, y_right_tile)

        def tile_to_lat(y_tile):
            n = float(2 ** zoom_level)
            lat_rad = np.arctan(np.sinh(np.pi * (1 - 2.0 * y_tile / n)))
            lat_deg = lat_rad * 180.0 / np.pi
            return lat_deg
        def tile_to_crs(x_tile, y_tile):
            n = float(2 ** zoom_level)
            lon_deg = x_tile / n * 360.0 - 180.0
            return (lon_deg, tile_to_lat(y_tile))
        def wrap_pi(lon):
            ret = lon
            while ret <= -180.0:
                ret += 360.0
            while ret >= 180.0:
                ret -= 360.0
            return ret

        crs_x_increment = 360.0 / (2**zoom_level)
        start_x_crs, start_y_crs = tile_to_crs(x_left_tile, y_left_tile)

        with tile_program as p:
            p.uniforms.crs_to_model = model_to_crs.inverse()
            p.uniforms.model_to_world = model_to_world
            p.uniforms.world_to_clip = world_to_clip
            p.uniforms.tile_image = int(0)

            for y_tile in range(y_left_tile, y_right_tile+1):

                y_crs = tile_to_lat(y_tile)
                crs_y_increment = tile_to_lat(y_tile+1) - y_crs
                x_crs = wrap_pi(start_x_crs)
                for x_tile in range(x_left_tile, x_right_tile+1):
                    tile_to_crs_mat = Matrix.product(
                        Matrix.translate((x_crs, y_crs)),
                        Matrix.scale((crs_x_increment, crs_y_increment, 1.0))
                    )
                    p.uniforms.tile_to_crs = tile_to_crs_mat

                    tile = self.get_tile(x_tile, y_tile, zoom_level)
                    if tile is not None:
                        tex = Qt.QOpenGLTexture(tile)
                        tex.create()
                        with Bound(tex, 0):
                            tile_mesh.draw()
                        tex.destroy()

                    x_crs += wrap_pi(crs_x_increment)



from .Schema              import Schema
from .ListSchema          import ListSchema, ListRoles, ListIconRoles
from .StringSchema        import StringSchema
from .IntSchema           import IntSchema, IntSliderSchema
from .FloatSchema         import FloatSliderSchema, FloatSchema
from .StructSchema        import StructSchema, MaybeStructSchema, StructRoles, SeparatorSchema
from .ColorSchema         import ColorSchema
from .UnionSchema         import UnionSchema
from .BoolSchema          import BoolSchema
from .ColorMapSchema      import ColorMapSchema
from .TileProviderSchema  import TileProviderSchema
from .DateSchema          import DateSchema

from .SchemaItemModel import SchemaItemModel, SchemaItemDelegate, SchemaView



from typing import Any, Optional, override
import datetime

from .Schema import Schema
from .__prelude__ import Qt

def _parse_q_date(q_date):
    year, month, day = q_date.getDate()
    return datetime.datetime.strptime(f"{year}-{month}-{day}", "%Y-%m-%d").timestamp()

class DateSchema(Schema):
    class DateNode(Schema.Node):
        def __init__(self, item_model, schema, model, name):
            super().__init__(item_model, schema)
            self.name = name
            self._dialog = None
            self.set_model(model)

        def createEditor(self, parent):
            self._editor = Qt.QCalendarWidget()
            self._editor.activated.connect(lambda q_date: self.requestModelChange.emit(_parse_q_date(q_date)))
            self._dialog = Qt.QDialog(parent)
            self._dialog.setWindowFlags(self._dialog.windowFlags() | Qt.Qt.WindowStaysOnTopHint)
            self._dialog.setWindowTitle(f"Choose a date for : {self.name}")
            layout = Qt.QVBoxLayout()
            layout.addWidget(self._editor)
            self._dialog.setLayout(layout)
            self._dialog.show()
            return self._dialog
        def destroyEditor(self):
            self._dialog.close()
            self._dialog = None
            return True
        def setModelData(self, __editor__):
            return True
        def setEditorData(self, editor):
            return True

        def data(self, role):
            if role == Qt.Qt.ItemDataRole.DisplayRole:
                date_str = datetime.datetime.strftime(datetime.datetime.fromtimestamp(self.model), "%Y-%m-%d")
                return f"{self.name} : {date_str}"
            return None

        def flags(self):
            return Qt.Qt.ItemFlag.ItemIsEditable | super().flags()
        def set_model(self, model):
            super().set_model(model)
            if self._dialog is not None:
                self.destroyEditor()
            self.dataChanged.emit([])

    def __init__(self, name):
        self.name = name

    def make_node(self, item_model, model):
        return self.DateNode(item_model, self, model, self.name)



from typing import Any

from .__prelude__ import Qt
from .Lens import SELF
from .Dynamic import Dynamic, PureDynamic

class dynamic_property:
    def __init__(self, initial = None):
        self.__initial = initial
    def __set_name__(self, klass, name):
        self.__name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        sig = instance.dynamic_attribute(self.name)
        return sig.value
    def __set__(self, instance, value):
        sig = instance.dynamic_attribute(self.name)
        sig.value = value
    def register(self, instance):
        return PureDynamic(self.__initial)
    @property
    def name(self):
        return self.__name

# Useful decorators
class CachedProperty(dynamic_property):
    class __CachedDynamic(Dynamic):
        def __init__(self, depends, func, instance, destroy):
            super().__init__()
            self.__depends = depends
            self.__destroy = destroy
            self.__func = func
            self.__last_computed = None
            self.__instance = instance
            self.__last_depend_triggered = None
            for dep in self.__depends:
                dep.value_changed.connect(self.__on_depend_changed(dep))

        def __on_depend_changed(self, dep):
            def doit():
                self.__last_depend_triggered = dep
                self.invalidate()
            return doit

        def invalidate(self):
            if hasattr(self, "_current_value"):
                del self._current_value
            self.value_changed.emit(self.__last_computed)
        def clear(self):
            self.invalidate()
            self.__destroy(self.__last_computed)
            self.__last_computed = None

        @property
        def value(self):
            if not hasattr(self, "_current_value"):
                self.__destroy(self.__last_computed)
                self._current_value = self.__func(self.__instance, self.__last_depend_triggered)
                self.__last_computed = self._current_value
            return self._current_value

    def __init__(self, depends, func, destroy):
        self.__depends = depends
        self.__func = func
        self.__destroy = destroy

    def register(self, instance):
        return self.__CachedDynamic([dep[instance] for dep in self.__depends], self.__func, instance, self.__destroy)
def cached_property(*depends, destroy = lambda _: None):
    return lambda f: CachedProperty(depends, lambda _self, _: f(_self), destroy)
cached_property.tracing = lambda *depends: lambda f: CachedProperty(depends, f)

def dynamic_method(*depends):
    def wrap(f):
        def ret(_self):
            def doit(*args, **kwargs):
                return f(_self, *args, **kwargs)
            return doit
        ret.__name__ = f.__name__
        return cached_property(*depends)(ret)
    return wrap

class view_property(dynamic_property):
    class __PrivateProperty:
        def __init__(self, view_name):
            self.__view_name = view_name
        def __get__(self, instance, owner):
            if instance is None:
                return self
            return instance.dynamic_attribute(self.__view_name).value
        def __set__(self, instance, value):
            instance.dynamic_attribute(self.__view_name).set_value(value)
    class __ReadonlyDynamic(Dynamic):
        def __init__(self, name):
            super().__init__()
            self.__value = None
            self.__name = name
        def set_value(self, value):
            old = self.__value
            self.__value = value
            self.value_changed.emit(old)
        @property
        def value(self):
            return self.__value
        @value.setter
        def value(self, v):
            raise Exception(f"view property '{self.__name}' is readonly")

    def __set_name__(self, klass, name):
        super().__set_name__(klass, name)
        self.__hidden_name = "_"+name
        setattr(klass, self.__hidden_name, self.__PrivateProperty(self.name))
    def register(self, instance):
        return self.__ReadonlyDynamic(self.name)

class derived_property(dynamic_property):
    def register(self, instance):
        ret = getattr(instance, "_dynamic_"+self.name, PureDynamic(None))
        return ret

class filtered_property(dynamic_property):
    class __FilteredDynamic(Dynamic):
        def __init__(self, instance, filt):
            super().__init__(None)
            self.__filter = filt
            self.__instance = instance
            self.__value = None
        @property
        def value(self):
            return self.__value
        @value.setter
        def value(self, v):
            old = self.__value
            self.__value = self.__filter(self.__instance, v)
            self.value_changed.emit(old)

    def __init__(self, filt):
        self.__filter = filt
    def register(self, instance):
        return self.__FilteredDynamic(instance, self.__filter)

class dynamic:
    variable           = dynamic_property
    filtered_variable  = filtered_property
    readonly           = view_property
    external           = derived_property
    memo               = cached_property
    method             = dynamic_method



import logging
import os
from .__prelude__ import Qt, INSARVIZ_ROOT


class LogsItemModel(Qt.QAbstractItemModel):
    def __init__(self, log_store):
        super().__init__()
        self._log_store = log_store
        self._log_store.new_record.connect(self._on_new_record)

    @Qt.Slot()
    def _on_new_record(self):
        size = len(self._log_store.records)
        self.beginInsertRows(Qt.QModelIndex(), size-1, size-1)
        self.endInsertRows()

    def rowCount(self, index):
        if index.isValid():
            return 0
        else:
            return len(self._log_store.records)
    def columnCount(self, index):
        return 4

    def parent(self, index):
        return Qt.QModelIndex()

    def headerData(self, section, orientation, role):
        if orientation == Qt.Qt.Horizontal:
            if role == Qt.Qt.ItemDataRole.DisplayRole:
                match section:
                    case 0:
                        return "Log level"
                    case 1:
                        return "Source"
                    case 2:
                        return "Function"
                    case 3:
                        return "Message"
        else:
            if role == Qt.Qt.ItemDataRole.DisplayRole:
                return str(section)
        return None

    def data(self, index, role):
        if index.isValid():
            record = self._log_store.records[index.row()]
            if role == Qt.Qt.ItemDataRole.DisplayRole:
                match index.column():
                    case 0:
                        return record.levelname
                    case 1:
                        return record.pathname.removeprefix(INSARVIZ_ROOT).removesuffix(".py").replace(os.sep, ".")+":"+str(record.lineno)
                    case 2:
                        return record.funcName
                    case 3:
                        return record.message

            if role == Qt.Qt.ItemDataRole.ForegroundRole:
                match record.levelno:
                    case logging.WARNING:
                        return Qt.QColor('orange')
                    case logging.ERROR:
                        return Qt.QColor('red')
                    case logging.DEBUG:
                        return Qt.QColor('grey')
                return None
            return None

    def index(self, row, column, parent):
        return self.createIndex(row, column, None)

class LoggerWidget(Qt.QTableView):
    def __init__(self, log_store):
        super().__init__()
        self.setModel(LogsItemModel(log_store))
        self.horizontalHeader().setStretchLastSection(True)
        self.verticalHeader().setSectionResizeMode(Qt.QHeaderView.ResizeMode.ResizeToContents)



import logging
import logging.handlers
import pathlib
import os
import multiprocessing
import warnings

logger = logging.getLogger('insarviz')

warnings.filterwarnings('ignore', r'All-NaN (slice|axis) encountered')

def _setup_logger(log_handler):
    from PySide6.QtCore import QObject, Signal

    class LogsStoreSignals(QObject):
        new_record = Signal()
    class LogsStore(logging.Handler):
        def __init__(self):
            super().__init__()
            self._records = []
            self._signals = LogsStoreSignals()

        @property
        def records(self):
            return self._records
        @property
        def new_record(self):
            return self._signals.new_record

        def emit(self, record):
            self._records.append(record)
            self.new_record.emit()

    log_messages = LogsStore()

    queue = multiprocessing.Queue()
    queue_handler = logging.handlers.QueueHandler(queue)

    queue_listener = logging.handlers.QueueListener(queue,
                                                    log_messages,
                                                    log_handler)

    class MPFilter:
        # Filter out messages from multiprocessing.Queue, to avoid deadlocks
        # See: https://docs.python.org/3/library/logging.handlers.html#queuehandler
        def filter(self, record):
            return not (record.levelno == logging.DEBUG and record.module == 'multiprocessing')

    logger.addFilter(MPFilter())
    logger.addHandler(queue_handler)

    return log_messages, queue_listener


INSARVIZ_ROOT = str(pathlib.Path(__file__).parent.parent) + os.sep

class InsarvizFilter:
    def filter(self, record):
        return record.name.startswith("insarviz")
logger.addFilter(InsarvizFilter())
logger.setLevel(logging.DEBUG)

class InsarvizFormatter(logging.Formatter):
    def __init__(self):
        super().__init__('%(asctime)s.%(msecs)03d %(levelname)s %(modulepath)s:%(lineno)d: %(message)s',
                         datefmt = '%Y-%m-%d %H:%M:%S')
    def format(self, record):
        record.modulepath = record.pathname.removeprefix(INSARVIZ_ROOT).removesuffix(".py").replace(os.sep, ".")
        return super().format(record)

log_handler = logging.StreamHandler()
log_handler.setFormatter(InsarvizFormatter())

log_messages, log_listener = _setup_logger(log_handler)
log_listener.start()



from ..__prelude__ import *

from ..shaders import Shader, PureShader
from ..misc import Qt, Matrix, Point, DEMTexture, Bound, ComputedValue, linmap, bresenham, Base10Increments, GLProgram, unit_square_to_image

def inOpenGLContext(function, context = None):
    if context is None:
        context = Qt.QOpenGLContext.globalShareContext()
    current = Qt.QOpenGLContext.currentContext()
    surface = current.surface()
    context.makeCurrent(surface)
    ret = function()
    context.functions().glFlush()
    if current is not None:
        current.makeCurrent(surface)
    return ret



import numpy as np

from .__prelude__ import (
    Qt, PureShader, Matrix, ColorMap, colormaps,
    dynamic, SELF,
    Dataset, destroy_texture
)
from .Layer import Layer, InitializedLayer

RASTER_SHADER = """
#version 400

uniform sampler2D {image};
uniform sampler1D {colormap};
uniform mat3 {geo_to_image};
uniform mat2 {value_to_color};
uniform float {opacity};
uniform float {height_scale};
uniform float {height_mean};

float {geo_height}(vec3 geo_coords) {{
   vec3 image_coords = {geo_to_image}*geo_coords;
   vec4 value = texture({image}, image_coords.xy / image_coords.z);
   return fma(value.x, {height_scale}, -{height_mean});
}}

vec4 {geo_color}(vec3 geo_coords, inout float up_opacity) {{
   vec3 image_coords = {geo_to_image}*geo_coords;
   vec4 value = texture({image}, image_coords.xy / image_coords.z);
   vec2 color_index = {value_to_color} * vec2(value.x, 1.0);
   vec4 color = texture({colormap}, color_index.x / color_index.y);
   return vec4(color.xyz, {opacity}*up_opacity*value.y);
}}
"""

class RasterLayer(Layer):
    name          = dynamic.variable("Raster Layer")
    colormap      = dynamic.variable()
    opacity       = dynamic.variable(1.0)
    is_enabled    = dynamic.variable(True)
    image         = dynamic.readonly()
    shaders       = dynamic.readonly()
    geo_to_image  = dynamic.readonly()
    dataset       = dynamic.readonly()
    band_number   = dynamic.readonly()
    dem_weight    = dynamic.variable(0.0)

    def __init__(self, dataset, band_number):
        super().__init__()
        self._dataset = dataset
        self._band_number = band_number
        self._image = dataset.read(band_number+1, masked=True)
        w,h = dataset.size
        self._geo_to_image = Matrix.product(
            dataset.pixel_to_crs,
            Matrix.scale((w,h,1))
        ).inverse()
        self._shaders = [
            RASTER_SHADER.format(
                **self.local_idents("image", "geo_to_image", "geo_color", "geo_height", "height_scale", "height_mean", "colormap", "opacity", "value_to_color")
            )
        ]
        self.colormap = ColorMap("grey", np.array(colormaps["grey"]) * 255.0)
        self.colormap.image_histogram = self.image_histogram

        self.dynamic_attribute("_redraw").value_changed.connect(lambda: self._redraw())

    @dynamic.method(SELF.colormap.value_to_color, SELF.colormap.createTexture, SELF.opacity, SELF.is_enabled, SELF.dem_weight)
    def _redraw(self):
        self.renderChanged.emit()

    @dynamic.memo(SELF.image)
    def image_histogram(self):
        img = self.image
        hist, bins = np.histogram(img.data[~img.mask], bins='fd', density=True)
        return (hist, bins)

    @dynamic.memo(SELF.image)
    def height_scale(self):
        a,b = np.nanquantile(self.image.data, 0.05, axis = None), np.nanquantile(self.image.data, 0.95, axis = None)
        return 1/max(abs(a), abs(b))
    @dynamic.memo(SELF.image, SELF.height_scale)
    def height_mean(self):
        return np.nanmean(self.image.data) * self.height_scale

    def GL_initialize(self, context):
        return InitializedRasterLayer(context, self)

    @property
    def description(self):
        return f"File: .../{self.dataset.file.name}"

    __mime_type__ = "x-application/insarviz/RasterLayer"
    @staticmethod
    def from_dict(dct, **__kwargs__):
        ret = RasterLayer(Dataset(dct["dataset"]), dct["band_number"])
        ret.opacity = dct['opacity']
        ret.is_enabled = dct.get('is_enabled', True)
        ret.colormap.init_from_dict(dct["colormap"])
        ret.dem_weight = dct.get('dem_weight', 0.0)
        ret.name = dct.get('name', "Raster Layer")
        return ret
    def to_dict(self):
        return {
            "opacity": self.opacity,
            "is_enabled": self.is_enabled,
            "dataset": str(self.dataset.file),
            "band_number": self.band_number,
            "colormap": self.colormap.to_dict(),
            "dem_weight": self.dem_weight,
            "name": self.name,
        }

class InitializedRasterLayer(InitializedLayer):
    def __init__(self, context, layer):
        super().__init__(context, layer)

    @dynamic.memo(SELF.layer.image,
                     destroy = destroy_texture)
    def GL_texture(self):
        texture = Qt.QOpenGLTexture(Qt.QOpenGLTexture.Target.Target2D)
        img = self.layer.image
        h, w = img.shape
        texture.setWrapMode(Qt.QOpenGLTexture.ClampToBorder)
        texture.setBorderColor(Qt.QColor(0,0,0,0))
        texture.setSize(w,h)
        texture.setFormat(Qt.QOpenGLTexture.TextureFormat.RG32F)
        texture.allocateStorage(Qt.QOpenGLTexture.PixelFormat.RG, Qt.QOpenGLTexture.PixelType.Float32)
        texture_data = np.stack([
            img.data,
            np.where(img.mask, 0.0, 1.0)
        ], axis=-1).astype(np.float32)
        texture.setData(Qt.QOpenGLTexture.PixelFormat.RG, Qt.QOpenGLTexture.PixelType.Float32, texture_data)
        texture.create()
        return texture

    @dynamic.memo(SELF.layer.colormap.createTexture,
                     destroy = destroy_texture)
    def GL_colormap_texture(self):
        return self.layer.colormap.createTexture()

    @dynamic.memo(SELF.GL_texture, SELF.GL_colormap_texture,
                  SELF.layer.opacity, SELF.layer.colormap.value_to_color,
                  SELF.layer.is_enabled, SELF.layer.height_scale, SELF.layer.dem_weight)
    def GL_setup(self):
        image_texture = self.GL_texture
        colormap_texture = self.GL_colormap_texture
        def doit(program):
            uni = self.local_uniforms(program.uniforms)
            uni.image           = image_texture
            uni.colormap        = colormap_texture
            uni.value_to_color  = self.layer.colormap.value_to_color
            uni.opacity         = self.layer.opacity if self.layer.is_enabled else 0.0
            uni.geo_to_image    = self.layer.geo_to_image
            uni.height_scale    = float(self.layer.height_scale * self.layer.dem_weight)
            uni.height_mean     = float(self.layer.height_mean * self.layer.dem_weight)
        return doit

    def GL_free(self):
        self.dynamic_attribute("GL_setup").clear()
        self.dynamic_attribute("GL_texture").clear()
        self.dynamic_attribute("GL_colormap_texture").clear()



from ..__prelude__ import *
from ..__logging__ import log_messages, INSARVIZ_ROOT

from ..misc import Qt, Matrix, Point, bresenham, ComputedValue, linmap, GLOBAL_THREAD_POOL, Runnable
from ..state import (
    ObservableList, ObservableStruct, WindowState, MaybeBandNumber,
    colormaps, Dataset,
    LinearCurve, SigmoidCurve, PeriodicCurve, HeavisideCurve, CurveSum, CurveParam, CurveEstimate,
    MapPoint, PointCurve, PointData,
    MapProfile, ProfileData, FocusPoint,
    MapScene, Scene, ProfileScene,
    GeoScene,
    Layer, SwipeLayer, PointsLayer, SelectedBandLayer, RasterLayer, RasterRGBLayer,
    SELF, EACH,
)

def color_icon(color):
    pix = Qt.QPixmap(16, 16)
    pix.fill(color)
    return Qt.QIcon(pix)



import math

class Base10Increments:
    MAJOR_INCREMENT = 0
    MINOR_INCREMENT = 1

    @classmethod
    def _gen_increments(cls,a,b):
        order = math.floor(math.log((b-a), 10))
        increment = (10.0**order)
        if (b-a) / increment < 5:
            order -= 1
            increment = (10.0**order)
        rng = (b-a)/increment
        start_n = math.ceil(a/increment)
        start = start_n*increment
        mod = start_n % 10
        count = math.ceil(rng)
        major_inc = 1
        if count >= 20:
            major_inc = 5
        elif count >= 10:
            major_inc = 2
        for i in range(count):
            inc_type = cls.MAJOR_INCREMENT if (mod+i) % major_inc == 0 else cls.MINOR_INCREMENT
            val = start+i*increment
            if val <= b:
                yield inc_type, val

    def __init__(self, a, b):
        if a <= b:
            self._a, self._b = a, b
        else:
            self._a, self._b = b, a

    def __iter__(self):
        return self._gen_increments(self._a, self._b)



from typing import Any
import numpy as np
import rasterio.windows
import warnings

from .__prelude__ import Qt, linmap, logger

from .observable import ObservableStruct, dynamic, SELF, Unique
from .curve      import Curve, LinearCurve, SigmoidCurve, PeriodicCurve, HeavisideCurve, LogCurve, ConstantCurve,CurveSum

def Enablable(SomeCurve):
    class Enablable(SomeCurve):
        is_enabled = dynamic.variable()
        def __init__(self, *args):
            super().__init__(*args)
            self.is_enabled = False

        def init_from_dict(self, dct):
            super().init_from_dict(dct)
            if "is_enabled" in dct:
                self.is_enabled = dct['is_enabled']

        def to_dict(self):
            return dict(super().to_dict(), is_enabled = self.is_enabled)
    return Enablable

class PointCurve(ObservableStruct):
    sigmoid_curve        = dynamic.readonly()
    linear_curve         = dynamic.readonly()
    seasonal_curve       = dynamic.readonly()
    semi_seasonal_curve  = dynamic.readonly()
    postseismic_curve    = dynamic.readonly()
    seism_timestamp      = dynamic.variable(0.0)
    timestamp_to_x       = dynamic.variable(lambda x: x)

    def __init__(self):
        super().__init__()
        self._sigmoid_curve        = Enablable(SigmoidCurve)()
        self._linear_curve         = Enablable(LinearCurve)()
        self._seasonal_curve       = Enablable(PeriodicCurve)()
        self._semi_seasonal_curve  = Enablable(PeriodicCurve)()
        self._postseismic_curve    = Enablable(HeavisideCurve)(CurveSum(LogCurve(), ConstantCurve()))
        self.dynamic_attribute("_set_seism_step").drive(lambda f: f())

    @dynamic.method(SELF.seism_timestamp, SELF.timestamp_to_x)
    def _set_seism_step(self):
        x = self.timestamp_to_x(self.seism_timestamp)
        logger.debug(f"Setting seism date to {x}")
        self.postseismic_curve.step = x

    __mime_type__ = "application/x-insarviz/PointCurve"
    def to_dict(self):
        return {
            "linear_curve": self.linear_curve.to_dict(),
            "sigmoid_curve": self.sigmoid_curve.to_dict(),
            "seasonal_curve": self.seasonal_curve.to_dict(),
            "semi_seasonal_curve": self.semi_seasonal_curve.to_dict(),
            "postseismic_curve": self.postseismic_curve.to_dict(),
            "seism_timestamp": self.seism_timestamp,
        }
    def init_from_dict(self, dct):
        self.linear_curve.init_from_dict(dct["linear_curve"])
        self.sigmoid_curve.init_from_dict(dct["sigmoid_curve"])
        if "seasonal_curve" in dct:
            self.seasonal_curve.init_from_dict(dct["seasonal_curve"])
        if "semi_seasonal_curve" in dct:
            self.semi_seasonal_curve.init_from_dict(dct["semi_seasonal_curve"])
        if "postseismic_curve" in dct:
            self.postseismic_curve.init_from_dict(dct["postseismic_curve"])
        if "seism_timestamp" in dct:
            self.seism_timestamp = dct["seism_timestamp"]

    def _curves(self):
        return [
            self.sigmoid_curve,
            self.linear_curve,
            self.seasonal_curve,
            self.semi_seasonal_curve,
            self.postseismic_curve,
        ]

    @dynamic.memo(
        SELF.sigmoid_curve.is_enabled,
        SELF.linear_curve.is_enabled,
        SELF.seasonal_curve.is_enabled,
        SELF.semi_seasonal_curve.is_enabled,
        SELF.postseismic_curve.is_enabled,
    )
    def is_empty(self):
        return not any([curve.is_enabled for curve in self._curves()])

    @dynamic.memo(
        SELF.is_empty,
        SELF.sigmoid_curve.initial_step,
        SELF.seasonal_curve.period,
        SELF.semi_seasonal_curve.period,
        SELF.postseismic_curve.step,
    )
    def full_curve(self):
        return CurveSum(*[curve for curve in self._curves() if curve.is_enabled])

class PointData(ObservableStruct):
    selected_band  = dynamic.readonly()
    point          = dynamic.variable()
    color          = dynamic.external()
    point_curve    = dynamic.external()
    show_in_plot   = dynamic.external()
    show_variance  = dynamic.external()

    def __init__(self, selected_band, point):
        super().__init__()
        self._selected_band = selected_band
        self._dynamic_color          = SELF.color[point]
        self._dynamic_point_curve    = SELF.point_curve[point]
        self._dynamic_show_in_plot   = SELF.show_in_plot[point]
        self._dynamic_show_variance  = SELF.show_variance[point]
        self.point = point

    def set_point(self, point):
        self.point = point

    @dynamic.memo(SELF.selected_band.dataset)
    def point_xs(self):
        return np.array(self.selected_band.dataset.band_timestamps)

    @dynamic.memo(SELF.selected_band.dataset, Unique(SELF.point.x), Unique(SELF.point.y), Unique(SELF.point.r))
    def point_ys(self):
        return self.__compute_point_values()

    @dynamic.memo(SELF.point_xs, SELF.point_ys)
    def point_absolute_values(self):
        mean, _ = self.point_ys
        return self.point_xs, mean

    @dynamic.memo(SELF.point_absolute_values, SELF.selected_band.reference_band_number)
    def point_values(self):
        ref = self.selected_band.reference_band_number
        xs,ys = self.point_absolute_values
        if ref is None:
            ref_val = 0
        else:
            ref_val = ys[ref]
        return xs, ys - ref_val

    @dynamic.memo(SELF.point_ys)
    def point_variance(self):
        _, var = self.point_ys
        return var

    @dynamic.memo(SELF.point_xs)
    def fromto_x(self):
        xs = self.point_xs
        return linmap(xs[0], xs[-1])

    @dynamic.memo(SELF.fromto_x, SELF.selected_band.dataset)
    def seasonal_period(self):
        if not self.selected_band.dataset.has_band_dates:
            return 1.0

        from_x, _ = self.fromto_x
        period = from_x(365*24*3600) - from_x(0)
        return period

    def __compute_point_values(self):
        p = self.point
        dataset = self.selected_band.dataset
        if p is None:
            zeros = np.zeros_like(np.array(dataset.band_timestamps))
            return zeros, zeros

        win = rasterio.windows.Window(p.x-p.r, p.y-p.r, 1+2*p.r, 1+2*p.r)
        # Create a mask of circular shape
        mask_row = np.reshape(np.arange(1+2*p.r), (1,1+2*p.r)) - p.r
        mask_col = np.reshape(np.arange(1+2*p.r), (1+2*p.r,1)) - p.r
        mask = np.where(mask_row*mask_row + mask_col*mask_col <= p.r*p.r, 1.0, np.nan)

        samples = dataset.read([i+1 for i in range(dataset.count)],
                               window = win, masked = True)
        masked_samples = np.reshape(np.where(samples.mask, np.nan, samples.data), (dataset.count, *mask.shape)) * mask

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore')
            return np.nanmean(masked_samples, axis=(1,2)), np.nanvar(masked_samples, axis=(1,2))

class MapPoint(ObservableStruct):
    name          = dynamic.variable()
    x             = dynamic.variable()
    y             = dynamic.variable()
    r             = dynamic.variable()
    r2            = dynamic.external() # Alias for 'r'
    color         = dynamic.variable()
    show_in_map   = dynamic.variable()
    show_in_plot  = dynamic.variable()
    show_variance = dynamic.variable(False)
    point_curve   = dynamic.variable()
    is_reference  = dynamic.variable(False)
    show_marker   = dynamic.variable(True)

    def __init__(self, name, x, y, r, color, show_in_map = True, show_in_plot = True, point_curve = None, show_variance = False, show_marker = True):
        super().__init__()
        self.name = name
        self.x = x
        self.y = y
        self._dynamic_r2 = self.dynamic_attribute("r")
        self.r = r
        self.color: Qt.QColor = color
        self.show_in_map = show_in_map
        self.show_in_plot = show_in_plot
        self.point_curve = point_curve if point_curve is not None else PointCurve()
        self.show_variance = show_variance
        self.show_marker = show_marker

    __mime_type__ = "x-application/insarviz/map-point"
    @staticmethod
    def from_dict(dct: dict):
        args = dict(dct, color = Qt.QColor(dct['color']))
        curve_dict = dct.get("point_curve", None)
        if curve_dict is not None:
            args["point_curve"] = PointCurve.from_dict(curve_dict)
        return MapPoint(**args)
    def to_dict(self):
        return {
            "name": self.name,
            "x": self.x,
            "y": self.y,
            "r": self.r,
            "color": self.color.name(),
            "show_in_map": self.show_in_map,
            "show_in_plot": self.show_in_plot,
            "show_variance": self.show_variance,
            "show_marker": self.show_marker,
            "point_curve": self.point_curve.to_dict()
        }

    @dynamic.method(SELF.color, SELF.x, SELF.y, SELF.r, SELF.show_in_map, SELF.is_reference)
    def draw(self, painter, opacity):
        if not self.show_in_map:
            return
        alpha = int(255 * opacity)
        if self.is_reference:
            pen = Qt.QPen(Qt.QColor("red"))
            pen.setWidth(5)
            painter.setPen(pen)
        else:
            if self.color.lightnessF() > 0.5:
                painter.setPen(Qt.QColor(0,0,0,alpha))
            else:
                painter.setPen(Qt.QColor(255, 255, 255, alpha))
        painter.setBrush(Qt.QColor(self.color.red(),self.color.green(),self.color.blue(),alpha))
        painter.drawEllipse(self.x-self.r+1,self.y-self.r+1,2*self.r-2,2*self.r-2)



import numpy as np
import math

from .__prelude__ import dynamic, SELF
from .CurveParam import CurveParam
from .Curve import Curve

class PeriodicCurve(Curve):
    phase      = dynamic.variable()
    amplitude  = dynamic.variable()
    period     = dynamic.variable(1.0)

    def __init__(self):
        super().__init__()
        self.phase = CurveParam("phase")
        self.amplitude = CurveParam("amplitude")
        self.amplitude.initial = 1.0
        self.period = 1.0

    __mime_type__ = "application/x-insarviz/PeriodicCurve"
    def init_from_dict(self, dct):
        super().init_from_dict(dct)
    def to_dict(self):
        return dict(super().to_dict(), period = self.period)

    @property
    def nparams(self):
        return 2
    @property
    def params(self):
        return [self.amplitude, self.phase]
    @dynamic.memo(SELF.period)
    def frequency(self):
        freq = 2.*math.pi / self.period
        return freq

    def func(self, xs, a, phi):
        return a*np.cos(self.frequency * xs + phi)
    def jacobian(self, xs, a, phi):
        xs_lin = self.frequency * xs + phi
        xcos = np.cos(xs_lin)
        xsin = np.sin(xs_lin)
        return np.array([xcos,-a*xsin])
    def hessian_epsilon(self, xs, err, grad, a, phi):
        xs_lin = self.frequency * xs + phi
        xcos = np.cos(xs_lin)
        xsin = np.sin(xs_lin)

        da_da = 0.0
        da_dphi = -np.sum(err*xsin)
        dphi_dphi = -a*np.sum(err*xcos)
        hessian = np.array([
            [da_da, da_dphi],
            [da_dphi, dphi_dphi]
        ])
        return np.dot(hessian @ grad, grad)



import numpy as np
from OpenGL import GL

from .__prelude__ import (
    Qt, ObservableStruct, ColorMap, colormaps, dynamic,
    Bound, inOpenGLContext, OpenGLPainter
)

class ExistingTexture:
    def __init__(self, texture_id):
        self.__texture_id = texture_id
    def destroy(self):
        arr = np.array([self.__texture_id])
        GL.glDeleteTextures(1, arr)

    def bind(self, index = 0):
        GL.glActiveTexture(int(GL.GL_TEXTURE0) + index)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.__texture_id)
    def release(self):
        pass
    def create(self):
        pass

class Layer(ObservableStruct):
    LAYER_ID = 0

    is_enabled = dynamic.variable(True)
    is_removable = True

    renderChanged = Qt.Signal()

    def __init__(self):
        super().__init__()
        self.__layer_id = self.LAYER_ID
        Layer.LAYER_ID += 1

    @dynamic.memo()
    def shaders(self):
        return []

    @property
    def layer_id(self):
        return self.__layer_id

    def local_ident(self, ident):
        return f"layer_{self.layer_id}_{ident}"
    def local_idents(self, *idents):
        return { ident: self.local_ident(ident) for ident in idents }

    def GL_initialize(self, context):
        raise NotImplementedError

class InitializedLayer(ObservableStruct):
    class __LocalUniforms:
        def __init__(self, layer, uni):
            self.__uniforms = uni
            self.__layer = layer
        def __setattr__(self, attr, val):
            if attr.startswith('_'):
                super().__setattr__(attr, val)
            else:
                setattr(self.__uniforms, self.__layer.local_ident(attr), val)

    layer = dynamic.readonly()
    context = dynamic.readonly()

    def __init__(self, context, layer):
        super().__init__()
        self._layer = layer
        self._context = context

    def local_uniforms(self, uni):
        return self.__LocalUniforms(self.layer, uni)

    def GL_create_shared_context(self):
        ret = Qt.QOpenGLContext()
        ret.setShareContext(self.context)
        ret.create()
        return ret

    def GL_paint_into_texture(self, w, h, shared_context, draw):
        fbo = Qt.QOpenGLFramebufferObject(w, h, Qt.QOpenGLFramebufferObject.Attachment.CombinedDepthStencil, target = GL.GL_TEXTURE_2D)
        def run():
            with Bound(fbo):
                gl = shared_context.functions()
                gl.glViewport(0,0,w,h)
                gl.glClearColor(0.0, 0.0, 0.0, 0.0)
                gl.glClear(GL.GL_COLOR_BUFFER_BIT)
                with OpenGLPainter(w,h) as painter:
                    draw(painter)
        inOpenGLContext(run, shared_context)
        return ExistingTexture(fbo.takeTexture())

    @dynamic.memo()
    def GL_setup(self):
        return lambda prog: None

    def GL_free(self):
        pass



import numpy as np
from OpenGL import GL
import math

from .__prelude__ import (
    SelectedBand, Qt, Shader, Point, Matrix, MapPoint, MapProfile, Bound,
    inOpenGLContext, OpenGLPainter,
    dynamic, SELF, EACH,
    GLProgram, PureShader,
    logger
)
from .Scene import Scene, SceneDrawing
from .layer import LayerStack
from .TileCache import TileCache

MAP_SHADER = """
#version 330
float {geo_height}(vec3 geo_coords);
vec4 {geo_color}(vec3 geo_coords, inout float up_opacity);

float geo_height(vec3 geo_coords) {{
  return {geo_height}(geo_coords);
}}
vec4 geo_color(vec3 geo_coords) {{
  float up_opacity = 1.0;
  return {geo_color}(geo_coords, up_opacity);
}}
"""

def location_path():
    path = Qt.QPainterPath()
    path.moveTo(38., 80.)
    path.cubicTo(35., 62., 26., 60., 20., 50.)
    path.cubicTo(0., 0., 80., 0., 60., 50.)
    path.cubicTo(54., 60., 48., 62., 45., 80.)
    path.lineTo(38., 80.)
    return path

class MapScene(Scene):
    selected_band      = dynamic.readonly()
    band_colormap      = dynamic.readonly()
    tile_cache         = dynamic.readonly()
    points             = dynamic.readonly()
    temp_point         = dynamic.variable(None)
    profiles           = dynamic.readonly()
    temp_profile       = dynamic.variable(None)
    overlay_layer      = dynamic.readonly()

    flip_horizontally  = dynamic.variable(False)
    flip_vertically    = dynamic.variable(False)
    geo_focus          = dynamic.variable(None)

    center             = dynamic.variable(Point(0.0, 0.0))
    heightUnits        = dynamic.variable(0.0)
    distance           = dynamic.variable(0.0)
    yaw                = dynamic.variable(0.0)
    pitch              = dynamic.variable(0.0)

    def __init__(self, selected_band: SelectedBand, points, profiles, layers, band_colormap):
        super().__init__()
        self._selected_band = selected_band
        self._shaders = [
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Vertex,"map/vertex.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.TessellationControl,"map/tess_control.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.TessellationEvaluation, "map/tess_eval.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.TessellationEvaluation, "map/utils.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment,"map/fragment.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment,"hsv_rgb.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment,"map/utils.glsl")
        ]
        self._tile_shaders = [
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Vertex,"tile/vertex.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment,"tile/fragment.glsl"),
        ]
        self._band_colormap = band_colormap

        self._tile_cache = TileCache()
        self.tile_cache.new_tile_available.connect(self.renderChanged.emit)

        self.color = (0.5, 0.5, 0.5, 1.0)

        self.distance = 5.0
        self._points = points
        self._point_counter = 1
        self._profiles = profiles
        self._profile_counter = 1

        self._overlay_layer = LayerStack(
            SELF.selected_band.dataset[self],
            SELF.temp_point[self],
            SELF.temp_profile[self],
            layers
        )
        self._overlay_layer.renderChanged.connect(self.renderChanged.emit)
        self.temp_point = None
        self.temp_profile = None

        self.dynamic_attribute("_rerender").value_changed.connect(lambda: self._rerender())

    def init_from_dict(self, dct):
        self.yaw = dct['yaw']
        self.pitch = dct['pitch']
        self.distance = dct['distance']
        self.center = Point(*dct['center'])
        self.heightUnits = dct['heightUnits']
        self.flip_horizontally = dct.get('flip_horizontally', False)
        self.flip_vertically = dct.get('flip_vertically', False)

    def to_dict(self):
        return {
            'yaw': self.yaw,
            'pitch': self.pitch,
            'distance': self.distance,
            'center': [self.center.x, self.center.y],
            'heightUnits': self.heightUnits,
            'flip_horizontally': self.flip_horizontally,
            'flip_vertically': self.flip_vertically,
        }

    def next_point_name(self):
        counter = self._point_counter
        self._point_counter += 1
        return f"Point {counter}"

    @dynamic.memo(SELF.distance, SELF.pitch, SELF.yaw, SELF.center)
    def model_to_world(self):
        return Matrix.product(
            Matrix.translate((0.0,0.0,self.distance)),
            Matrix.rotate3(self.pitch, (-1.0, 0.0, 0.0, 0.0)),
            Matrix.rotate3(self.yaw, (0.0, 0.0, 1.0, 0.0)),
            Matrix.translate((self.center.x, self.center.y, 0.0))
        )
    @dynamic.memo(SELF.flip_horizontally, SELF.flip_vertically)
    def model_to_texture(self):
        h_flip = -1.0 if self.flip_horizontally else 1.0
        v_flip = -1.0 if self.flip_vertically else 1.0
        return Matrix.product(
            Matrix.scale((h_flip, v_flip, 1.0)),
            Matrix.scale((0.5,0.5,1.0)),
            Matrix.translate((h_flip,v_flip))
        )

    @dynamic.memo(SELF.band_colormap.xzero, SELF.band_colormap.xone)
    def value_to_elevation(self):
        lower, upper = self.band_colormap.xzero, self.band_colormap.xone
        scale = max(abs(lower), abs(upper))
        return Matrix.scale((1./scale, 1.0))
    @dynamic.memo(SELF.selected_band.dataset)
    def image_to_geo(self):
        w,h = self.selected_band.size
        return Matrix.product(
            self.selected_band.dataset.pixel_to_crs,
            Matrix.scale((w,h,1))
        )

    @dynamic.memo(SELF.selected_band.dataset)
    def show_background(self):
        return self.selected_band.dataset.is_georeferenced

    @dynamic.method(SELF.selected_band.image, SELF.selected_band.reference_image,
                    SELF.heightUnits, SELF.model_to_world, SELF.temp_point, SELF.temp_profile,
                    SELF.model_to_texture, SELF.geo_focus,
                    SELF.band_colormap.image, SELF.band_colormap.value_to_color,
                    SELF.points / EACH(SELF.show_marker),
                    SELF.profiles / EACH(SELF.show_focus_marker))
    def _rerender(self):
        self.renderChanged.emit()

    def create_point(self, dragStart: tuple[int, int], dragEnd: tuple[int, int]):
        xa,ya = dragStart
        xb,yb = dragEnd
        model_to_image = Matrix.product(
            self.selected_band.texture.texture_to_image,
            self.model_to_texture
        )
        def sq(x):
            return x*x
        dx_image, dy_image, _ = model_to_image.transform_vect((xb-xa, yb-ya, 0.0))
        w,h = self.selected_band.size
        # Clamp the radius to 100 pixels, to avoid users accidentally
        # computing the mean of the whole dataset
        radius = min(int(math.sqrt(sq(dx_image*w) + sq(dy_image*h))), 100)
        x_image, y_image = model_to_image.transform_point(dragStart)
        return MapPoint("Point", x_image*w, y_image*h, radius, Qt.QColor("blue"))
    def create_profile(self, dragStart):
        dragEnd = dragStart
        x0, y0 = self.__model_to_pixel(*dragStart)
        x1, y1 = self.__model_to_pixel(*dragEnd)
        counter = self._profile_counter
        self._profile_counter = counter + 1
        return MapProfile(f"Profile {counter}", 10, Qt.QColor("orange"), True, x0, y0, [(x1, y1)])
    def set_last_temp_profile_point(self, position):
        self.temp_profile.end_points[-1] = self.__model_to_pixel(*position)
        self.dynamic_attribute("temp_profile").value_changed.emit(self.temp_profile)
    def append_temp_profile_point(self, position):
        new_point = self.__model_to_pixel(*position)
        if len(self.temp_profile.end_points) < 2 or new_point != self.temp_profile.end_points[-2]:
            self.temp_profile.end_points.append(new_point)
        self.dynamic_attribute("temp_profile").value_changed.emit(self.temp_profile)

    def __model_to_pixel(self, x_model, y_model):
        model_to_image = Matrix.product(
            self.selected_band.texture.texture_to_image,
            self.model_to_texture
        )
        x_image, y_image = model_to_image.transform_point((x_model, y_model))
        w , h = self.selected_band.size
        return (int(x_image*w), int(y_image*h))

    def initializeDrawing(self, context):
        gl = context.functions()
        logger.debug("Loading map scene (max texture size : %d)", gl.glGetIntegerv(GL.GL_MAX_TEXTURE_SIZE))
        return InitializedMapScene(context, self)

class InitializedMapScene(SceneDrawing["MapScene"]):
    class IdleMode(SceneDrawing.IdleMode):
        def on_mouse_wheel(self, position, angle_x, angle_y):
            zoom = angle_y / 32.0
            def doit():
                self.drawing.distance *= math.exp(math.log(0.25)*math.radians(zoom))
            self.initialized_drawing.preserving_coords(position, doit)
            return self

        def on_mouse_button_pressed(self, position, button):
            if button == Qt.Qt.MouseButton.LeftButton:
                return self.initialized_drawing.DraggingMode(self.initialized_drawing, position)
            return self

        def on_key_pressed(self, button):
            if button == Qt.Qt.Key_Control:
                return self.initialized_drawing.ControlMode(self.initialized_drawing)
            elif button == Qt.Qt.Key_Shift:
                return self.initialized_drawing.ShiftMode(self.initialized_drawing)
            elif button == Qt.Qt.Key_Right or button == Qt.Qt.Key_Up:
                self.drawing.selected_band.band_number += 1
                return self
            elif button == Qt.Qt.Key_Left or button == Qt.Qt.Key_Down:
                self.drawing.selected_band.band_number -= 1
                return self
            return self

    class DraggingMode(SceneDrawing.IdleMode):
        def __init__(self, initialized_map_scene, start_pos):
            super().__init__(initialized_map_scene)
            self.__start_position = start_pos

        def on_mouse_move(self, new_pos):
            xa,ya = self.initialized_drawing.clip_to_model(*self.__start_position)
            xb,yb = self.initialized_drawing.clip_to_model(*new_pos)
            self.drawing.center += Point(xb-xa, yb-ya)
            return self.initialized_drawing.DraggingMode(self.initialized_drawing, new_pos)

        def on_mouse_button_released(self, pos, button):
            return self.initialized_drawing.IdleMode(self.initialized_drawing)

    class ControlMode(SceneDrawing.IdleMode):
        def on_mouse_wheel(self, position, angle_x, angle_y):
            angle = angle_y / 32.0
            def doit():
                self.drawing.yaw += angle
            self.initialized_drawing.preserving_coords(position, doit)
            return self

        def on_key_released(self, button):
            return self.initialized_drawing.IdleMode(self.initialized_drawing)
        def on_key_pressed(self, key):
            if key == Qt.Qt.Key_Shift:
                return self.initialized_drawing.ControlShiftMode(self.initialized_drawing)
            return self

        def on_mouse_button_pressed(self, position, button):
            if button == Qt.Qt.MouseButton.LeftButton:
                return self.initialized_drawing.CreatePointMode(self.initialized_drawing, position, self.__class__)
            return self

    class CreatePointMode(SceneDrawing.IdleMode):
        def __init__(self, initialized_scene, start_pos, return_mode):
            super().__init__(initialized_scene)
            self.__start_position = start_pos
            self.__return_mode = return_mode

        def on_mouse_move(self, position):
            model_center = self.initialized_drawing.clip_to_model(*self.__start_position)
            model_edge = self.initialized_drawing.clip_to_model(*position)
            self.drawing.temp_point = self.drawing.create_point(model_center, model_edge)
            return self

        def on_mouse_button_released(self, position, button):
            model_center = self.initialized_drawing.clip_to_model(*self.__start_position)
            model_edge = self.initialized_drawing.clip_to_model(*position)
            point = self.drawing.create_point(model_center, model_edge)
            point.name = self.drawing.next_point_name()
            self.drawing.points.append(point)
            self.drawing.temp_point = None
            return self.__return_mode(self.initialized_drawing)

    class ShiftMode(SceneDrawing.IdleMode):
        def on_mouse_wheel(self, position, angle_x, angle_y):
            angle = angle_y / 32.0
            def doit():
                self.drawing.pitch = max(0.0,min(85.0,self.drawing.pitch + angle))
            self.initialized_drawing.preserving_coords(position, doit)
            return self

        def on_key_released(self, button):
            return self.initialized_drawing.IdleMode(self.initialized_drawing)
        def on_key_pressed(self, key):
            if key == Qt.Qt.Key_Control:
                return self.initialized_drawing.ControlShiftMode(self.initialized_drawing)
            return self

        def on_mouse_button_pressed(self, position, button):
            if button == Qt.Qt.MouseButton.LeftButton:
                return self.initialized_drawing.CreateProfileMode(self.initialized_drawing, position)
            return self

    class CreateProfileMode(SceneDrawing.IdleMode):
        def __init__(self, initialized_scene, position):
            super().__init__(initialized_scene)
            self.drawing.temp_profile = self.drawing.create_profile(self.initialized_drawing.clip_to_model(*position))

        def __finalize_profile(self):
            profile = self.drawing.temp_profile
            self.drawing.temp_profile = None
            if len(profile.end_points) > 1:
                profile.end_points[-1:] = []
                self.drawing.profiles.append(profile)
            return self.initialized_drawing.IdleMode(self.initialized_drawing)

        def on_key_released(self, key):
            if key == Qt.Qt.Key_Shift:
                return self.__finalize_profile()
            return self

        def on_mouse_button_released(self, position, button):
            if button == Qt.Qt.MouseButton.LeftButton:
                self.drawing.append_temp_profile_point(self.initialized_drawing.clip_to_model(*position))
            return self
        def on_mouse_button_pressed(self, position, button):
            if button == Qt.Qt.MouseButton.LeftButton:
                self.drawing.append_temp_profile_point(self.initialized_drawing.clip_to_model(*position))
            if button == Qt.Qt.MouseButton.RightButton:
                return self.__finalize_profile()
            return self

        def on_mouse_wheel(self, position, angle_x, angle_y):
            self.drawing.temp_profile.r = max(0, self.drawing.temp_profile.r + int(angle_y / 64.0))
            self.drawing.dynamic_attribute("temp_profile").value_changed.emit(self.drawing.temp_profile)
            return self

        def on_mouse_move(self, position):
            self.drawing.set_last_temp_profile_point(self.initialized_drawing.clip_to_model(*position))
            return self

    class ControlShiftMode(SceneDrawing.IdleMode):
        def on_mouse_wheel(self, position, angle_x, angle_y):
            angle = angle_y / 32.0
            def doit():
                self.drawing.heightUnits = min(1.0, max(0, self.drawing.heightUnits + angle/360.0))
            self.initialized_drawing.preserving_coords(position, doit)
            return self

        def on_key_released(self, button):
            if button == Qt.Qt.Key_Control:
                return self.initialized_drawing.ShiftMode(self.initialized_drawing)
            elif button == Qt.Qt.Key_Shift:
                return self.initialized_drawing.ControlMode(self.initialized_drawing)
            return self

    def __init__(self, context, map_scene):
        super().__init__(context, map_scene)

        self.__tile_program: Qt.QOpenGLShaderProgram = GLProgram(map_scene._tile_shaders)
        self.__tile_mesh = self.__tile_program.create_square_mesh(GL.GL_TRIANGLES, split=0, texture_to_model = Matrix.identity(3))

        self.__initialized_overlay = self.drawing.overlay_layer.GL_initialize(context)
        GL.glPatchParameteri(GL.GL_PATCH_VERTICES, 3)

    def idle_mode(self):
        return self.IdleMode(self)

    def preserving_coords(self, coords, doit):
        x_b, y_b = self.clip_to_model(*coords)
        doit()
        x_a, y_a = self.clip_to_model(*coords)
        self.drawing.center += Point(x_a - x_b, y_a - y_b)

    @dynamic.memo(SELF.drawing.overlay_layer.shaders)
    def GL_program(self):
        return GLProgram(self.drawing._shaders + [
            PureShader(Qt.QOpenGLShader.ShaderTypeBit.Fragment, shader)
            for shader in self.drawing.overlay_layer.shaders
        ] + [
            PureShader(Qt.QOpenGLShader.ShaderTypeBit.TessellationEvaluation, shader)
            for shader in self.drawing.overlay_layer.shaders
        ] + [
            PureShader(Qt.QOpenGLShader.ShaderTypeBit.Fragment, MAP_SHADER.format(
                geo_color = self.drawing.overlay_layer.local_ident("geo_color"),
                geo_height = self.drawing.overlay_layer.local_ident("geo_height")
            )),
            PureShader(Qt.QOpenGLShader.ShaderTypeBit.TessellationEvaluation, MAP_SHADER.format(
                geo_color = self.drawing.overlay_layer.local_ident("geo_color"),
                geo_height = self.drawing.overlay_layer.local_ident("geo_height")
            ))

        ])
    @dynamic.memo(SELF.GL_program, SELF.drawing.model_to_texture)
    def GL_mesh(self):
        return self.GL_program.create_square_mesh(
            GL.GL_PATCHES, split=8,
            texture_to_model = self.drawing.model_to_texture.inverse())

    @dynamic.memo(SELF.viewport_size)
    def world_to_clip(self):
        w,h = self.viewport_size
        hfactor = max(1.0, w/h)/4.0
        vfactor = max(1.0, h/w)/4.0
        near = 4.0*0.005
        return Matrix.frustum(-hfactor*near, hfactor*near,
                              -vfactor*near, vfactor*near,
                              -near, -1000.0)
    @dynamic.memo(SELF.viewport_size)
    def tiles_per_screen(self):
        w,h = self.viewport_size
        return max(w/256.0, h/256.0) # Each tile is 256 pixels wide
    @dynamic.memo(SELF.viewport_size)
    def clip_to_pixel(self):
        w,h = self.viewport_size
        return Matrix.scale((w/2.0, -h/2.0, 1.0)) * Matrix.translate((1.0,-1.0))

    def paintGL(self):
        program = self.GL_program
        mesh = self.GL_mesh
        glfunc = self.context.functions()
        setup_overlay = self.__initialized_overlay.GL_setup

        glfunc.glEnable(GL.GL_DEPTH_TEST)

        glfunc.glClearColor(*self.drawing.color)
        glfunc.glClear(int(GL.GL_COLOR_BUFFER_BIT) | int(GL.GL_DEPTH_BUFFER_BIT))

        if self.drawing.show_background:
            if self.drawing.pitch < 60.0:
                self.drawing.tile_cache.GL_paint_background(
                    glfunc,
                    model_to_crs = Matrix.product(
                        self.drawing.selected_band.image_to_crs,
                        self.drawing.selected_band.texture.texture_to_image,
                        self.drawing.model_to_texture
                    ),
                    clip_to_model = self.clip_to_model,
                    model_to_world = self.drawing.model_to_world,
                    world_to_clip = self.world_to_clip,
                    tiles_per_screen = self.tiles_per_screen,
                )

        dem_texture = self.drawing.selected_band.texture
        ref_texture = self.drawing.selected_band.reference_texture
        tex_to_image = dem_texture.texture_to_image

        glfunc.glEnable(GL.GL_BLEND)
        glfunc.glBlendFuncSeparate(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA, GL.GL_ONE, GL.GL_ONE)
        with program as p:
            p.uniforms.image = dem_texture
            p.uniforms.ref_image = ref_texture

            p.uniforms.texture_to_image = tex_to_image
            p.uniforms.image_to_texture = tex_to_image.inverse()

            p.uniforms.image_to_geo = self.drawing.image_to_geo
            p.uniforms.geo_to_model = Matrix.product(
                self.drawing.image_to_geo,
                tex_to_image,
                self.drawing.model_to_texture,
            ).inverse()
            setup_overlay(p)

            p.uniforms.heightUnits = float(self.drawing.heightUnits)
            p.uniforms.value_to_elevation = self.drawing.value_to_elevation
            _,_,z_near = self.world_to_clip.inverse().transform_point((0.0,0.0,-1.0))
            p.uniforms.near = z_near

            p.uniforms.world_to_clip = self.world_to_clip
            p.uniforms.model_to_world = self.drawing.model_to_world

            mesh.draw()

        with OpenGLPainter(*self.viewport_size) as painter:
            if self.drawing.geo_focus is not None:
                center = (tex_to_image * self.drawing.model_to_texture).inverse().transform_point(self.drawing.geo_focus)
                clip_cx, clip_cy, _ = (self.world_to_clip * self.drawing.model_to_world).transform_point((*center, 0.0))
                pixel_cx, pixel_cy = self.clip_to_pixel.transform_point((clip_cx, -clip_cy))
                painter.setBrush(Qt.QColor('red'))
                painter.drawEllipse(Qt.QPointF(pixel_cx, pixel_cy), 10 , 10)

            marker_path = location_path()
            path_x = marker_path.boundingRect().center().x()
            path_y = marker_path.boundingRect().bottom()

            w,h = self.drawing.selected_band.size

            point_to_model = Matrix.product(
                self.drawing.model_to_texture.inverse(),
                tex_to_image.inverse(),
                Matrix.scale((1/w, 1/h, 1)),
            )
            model_to_clip = Matrix.product(
                self.world_to_clip,
                self.drawing.model_to_world,
            )
            def marker_clip(point_x, point_y):
                model_x, model_y = point_to_model.transform_point((point_x, point_y))
                return model_to_clip.transform_point((model_x, model_y, 0.0))
            def show_marker_at(color, clip_x, clip_y):
                pixel_x, pixel_y = self.clip_to_pixel.transform_point((clip_x, -clip_y))
                painter.setBrush(color)
                painter.save()
                painter.translate(pixel_x, pixel_y)
                painter.scale(1,-1)
                painter.translate(-path_x,-path_y)
                painter.drawPath(marker_path)
                painter.restore()

            markers = []
            for point in self.drawing.points:
                if point.show_marker:
                    markers.append((point.color, marker_clip(point.x, point.y)))

            for profile in self.drawing.profiles:
                if profile.show_focus_marker:
                    markers.append((profile.color, marker_clip(*profile.focus_center)))

            def marker_z(marker):
                match marker:
                    case _, coords:
                        return coords[2]
            for color, coords in sorted(markers, key=marker_z, reverse=True):
                clip_x, clip_y, _ = coords
                show_marker_at(color, clip_x, clip_y)

    def freeGL(self):
        self.__initialized_overlay.GL_free()
